"""
PC-BASIC - video_sdl2.py
Graphical interface based on PySDL2

(c) 2015--2023 Rob Hagemans
This file is released under the GNU GPL version 3 or later.
"""

import logging
import ctypes
import os
import sys
import functools
import operator
from contextlib import contextmanager

from ..compat import iteritems, unichr
from ..compat import WIN32, BASE_DIR, PLATFORM, PY2
from ..compat import set_dpi_aware

from .base import EnvironmentCache
from .base import video_plugins, InitFailed, NOKILL_MESSAGE
from ..basic.base import signals
from ..basic.base import scancode
from ..basic.base import bytematrix
from ..basic.base.eascii import as_unicode as uea
from .video import VideoPlugin
from . import window
from . import clipboard



###############################################################################
# video settings

# refresh cycle parameters
# number of cycles to change blink state
BLINK_CYCLES = 5
# number of distinct blink states
N_BLINK_STATES = 4
# ms duration of a blink
BLINK_TIME = 120
CYCLE_TIME = BLINK_TIME // BLINK_CYCLES


# blank icon
BLANK_ICON = ((0,) * 16) * 16


###############################################################################
# locate and load SDL libraries

if False:
    # packagers take note, we need these imports to be bundled
    from . import sdl2loader
    from . import sdl2

sdl2 = None

# custom dll location
LIB_DIR = os.path.join(BASE_DIR, 'lib')


def _import_sdl2():
    """Import the sdl2 bindings and define constants."""
    global sdl2
    global SCAN_TO_SCAN, ALT_SCAN_TO_EASCII, MOD_TO_SCAN
    global KEY_TO_EASCII, SHIFT_KEY_TO_EASCII, CTRL_KEY_TO_EASCII

    # look for SDL2.dll / libSDL2.dylib / libSDL2.so:
    # first in LIB_DIR, then in pysdl2dll module, then the standard search path
    # this means that user should not have dll in LIB_DIR if they want to use another one
    from . import sdl2loader
    sdl2loader.load_dlls(LIB_DIR)

    # this raises ImportError if no library has been found
    from . import sdl2


    ###############################################################################
    # keyboard codes

    # these are PC keyboard scancodes
    SCAN_TO_SCAN = {
        # top row
        sdl2.SDL_SCANCODE_ESCAPE: scancode.ESCAPE, sdl2.SDL_SCANCODE_1: scancode.N1,
        sdl2.SDL_SCANCODE_2: scancode.N2, sdl2.SDL_SCANCODE_3: scancode.N3,
        sdl2.SDL_SCANCODE_4: scancode.N4, sdl2.SDL_SCANCODE_5: scancode.N5,
        sdl2.SDL_SCANCODE_6: scancode.N6, sdl2.SDL_SCANCODE_7: scancode.N7,
        sdl2.SDL_SCANCODE_8: scancode.N8, sdl2.SDL_SCANCODE_9: scancode.N9,
        sdl2.SDL_SCANCODE_0: scancode.N0, sdl2.SDL_SCANCODE_MINUS: scancode.MINUS,
        sdl2.SDL_SCANCODE_EQUALS: scancode.EQUALS, sdl2.SDL_SCANCODE_BACKSPACE: scancode.BACKSPACE,
        # row 1
        sdl2.SDL_SCANCODE_TAB: scancode.TAB, sdl2.SDL_SCANCODE_Q: scancode.q,
        sdl2.SDL_SCANCODE_W: scancode.w, sdl2.SDL_SCANCODE_E: scancode.e,
        sdl2.SDL_SCANCODE_R: scancode.r, sdl2.SDL_SCANCODE_T: scancode.t,
        sdl2.SDL_SCANCODE_Y: scancode.y, sdl2.SDL_SCANCODE_U: scancode.u,
        sdl2.SDL_SCANCODE_I: scancode.i, sdl2.SDL_SCANCODE_O: scancode.o,
        sdl2.SDL_SCANCODE_P: scancode.p, sdl2.SDL_SCANCODE_LEFTBRACKET: scancode.LEFTBRACKET,
        sdl2.SDL_SCANCODE_RIGHTBRACKET: scancode.RIGHTBRACKET,
        sdl2.SDL_SCANCODE_RETURN: scancode.RETURN, sdl2.SDL_SCANCODE_KP_ENTER: scancode.RETURN,
        # row 2
        sdl2.SDL_SCANCODE_RCTRL: scancode.CTRL, sdl2.SDL_SCANCODE_LCTRL: scancode.CTRL,
        sdl2.SDL_SCANCODE_A: scancode.a, sdl2.SDL_SCANCODE_S: scancode.s,
        sdl2.SDL_SCANCODE_D: scancode.d, sdl2.SDL_SCANCODE_F: scancode.f,
        sdl2.SDL_SCANCODE_G: scancode.g, sdl2.SDL_SCANCODE_H: scancode.h,
        sdl2.SDL_SCANCODE_J: scancode.j, sdl2.SDL_SCANCODE_K: scancode.k,
        sdl2.SDL_SCANCODE_L: scancode.l, sdl2.SDL_SCANCODE_SEMICOLON: scancode.SEMICOLON,
        sdl2.SDL_SCANCODE_APOSTROPHE: scancode.QUOTE, sdl2.SDL_SCANCODE_GRAVE: scancode.BACKQUOTE,
        # row 3
        sdl2.SDL_SCANCODE_LSHIFT: scancode.LSHIFT, sdl2.SDL_SCANCODE_BACKSLASH: scancode.BACKSLASH,
        sdl2.SDL_SCANCODE_Z: scancode.z, sdl2.SDL_SCANCODE_X: scancode.x,
        sdl2.SDL_SCANCODE_C: scancode.c, sdl2.SDL_SCANCODE_V: scancode.v,
        sdl2.SDL_SCANCODE_B: scancode.b, sdl2.SDL_SCANCODE_N: scancode.n,
        sdl2.SDL_SCANCODE_M: scancode.m, sdl2.SDL_SCANCODE_COMMA: scancode.COMMA,
        sdl2.SDL_SCANCODE_PERIOD: scancode.PERIOD, sdl2.SDL_SCANCODE_SLASH: scancode.SLASH,
        sdl2.SDL_SCANCODE_RSHIFT: scancode.RSHIFT, sdl2.SDL_SCANCODE_SYSREQ: scancode.SYSREQ,
        sdl2.SDL_SCANCODE_LALT: scancode.ALT,
        # don't catch right-Alt as it may inhibit AltGr on Windows
        # sdl2.SDL_SCANCODE_RALT: scancode.ALT,
        sdl2.SDL_SCANCODE_SPACE: scancode.SPACE, sdl2.SDL_SCANCODE_CAPSLOCK: scancode.CAPSLOCK,
        # function keys
        sdl2.SDL_SCANCODE_F1: scancode.F1, sdl2.SDL_SCANCODE_F2: scancode.F2,
        sdl2.SDL_SCANCODE_F3: scancode.F3, sdl2.SDL_SCANCODE_F4: scancode.F4,
        sdl2.SDL_SCANCODE_F5: scancode.F5, sdl2.SDL_SCANCODE_F6: scancode.F6,
        sdl2.SDL_SCANCODE_F7: scancode.F7, sdl2.SDL_SCANCODE_F8: scancode.F8,
        sdl2.SDL_SCANCODE_F9: scancode.F9, sdl2.SDL_SCANCODE_F10: scancode.F10,
        sdl2.SDL_SCANCODE_F11: scancode.F11, sdl2.SDL_SCANCODE_F12: scancode.F12,
        # top of keypad
        sdl2.SDL_SCANCODE_NUMLOCKCLEAR: scancode.NUMLOCK,
        sdl2.SDL_SCANCODE_SCROLLLOCK: scancode.SCROLLOCK, sdl2.SDL_SCANCODE_PAUSE: scancode.BREAK,
        # keypad
        sdl2.SDL_SCANCODE_KP_MULTIPLY: scancode.KPTIMES,
        sdl2.SDL_SCANCODE_PRINTSCREEN: scancode.PRINT,
        sdl2.SDL_SCANCODE_KP_7: scancode.KP7, sdl2.SDL_SCANCODE_HOME: scancode.HOME,
        sdl2.SDL_SCANCODE_KP_8: scancode.KP8, sdl2.SDL_SCANCODE_UP: scancode.UP,
        sdl2.SDL_SCANCODE_KP_9: scancode.KP9, sdl2.SDL_SCANCODE_PAGEUP: scancode.PAGEUP,
        sdl2.SDL_SCANCODE_KP_MINUS: scancode.KPMINUS, sdl2.SDL_SCANCODE_KP_4: scancode.KP4,
        sdl2.SDL_SCANCODE_LEFT: scancode.LEFT, sdl2.SDL_SCANCODE_KP_5: scancode.KP5,
        sdl2.SDL_SCANCODE_KP_6: scancode.KP6, sdl2.SDL_SCANCODE_RIGHT: scancode.RIGHT,
        sdl2.SDL_SCANCODE_KP_PLUS: scancode.KPPLUS, sdl2.SDL_SCANCODE_KP_1: scancode.KP1,
        sdl2.SDL_SCANCODE_END: scancode.END, sdl2.SDL_SCANCODE_KP_2: scancode.KP2,
        sdl2.SDL_SCANCODE_DOWN: scancode.DOWN, sdl2.SDL_SCANCODE_KP_3: scancode.KP3,
        sdl2.SDL_SCANCODE_PAGEDOWN: scancode.PAGEDOWN,
        sdl2.SDL_SCANCODE_KP_0: scancode.KP0, sdl2.SDL_SCANCODE_INSERT: scancode.INSERT,
        sdl2.SDL_SCANCODE_KP_PERIOD: scancode.KPPOINT, sdl2.SDL_SCANCODE_DELETE: scancode.DELETE,
        # extensions
        sdl2.SDL_SCANCODE_NONUSBACKSLASH: scancode.INT1,
        # windows keys
        sdl2.SDL_SCANCODE_LGUI: scancode.LSUPER, sdl2.SDL_SCANCODE_RGUI: scancode.RSUPER,
        sdl2.SDL_SCANCODE_MENU: scancode.MENU,
        # Japanese keyboards
        # mapping to SDL scancodes unknown
        #HIRAGANA_KATAKANA = 0x70
        # backslash/underscore on Japanese keyboard
        #INT3 = 0x73
        #HENKAN = 0x79
        # this is a guess based on https://wiki.libsdl.org/SDL_Scancode
        # and http://www.quadibloc.com/comp/scan.htm
        sdl2.SDL_SCANCODE_LANG4: scancode.MUHENKAN,
        sdl2.SDL_SCANCODE_LANG5: scancode.ZENKAKU_HANKAKU,
        sdl2.SDL_SCANCODE_INTERNATIONAL3: scancode.INT4,
        # Korean keyboards
        sdl2.SDL_SCANCODE_LANG2: scancode.HANJA,
        sdl2.SDL_SCANCODE_LANG1: scancode.HAN_YEONG,
    }

    KEY_TO_EASCII = {
        sdl2.SDLK_F1: uea.F1, sdl2.SDLK_F2: uea.F2, sdl2.SDLK_F3: uea.F3, sdl2.SDLK_F4: uea.F4,
        sdl2.SDLK_F5: uea.F5, sdl2.SDLK_F6: uea.F6, sdl2.SDLK_F7: uea.F7, sdl2.SDLK_F8: uea.F8,
        sdl2.SDLK_F9: uea.F9, sdl2.SDLK_F10: uea.F10, sdl2.SDLK_F11: uea.F11,
        sdl2.SDLK_F12: uea.F12, sdl2.SDLK_HOME: uea.HOME, sdl2.SDLK_UP: uea.UP,
        sdl2.SDLK_PAGEUP: uea.PAGEUP, sdl2.SDLK_LEFT: uea.LEFT, sdl2.SDLK_RIGHT: uea.RIGHT,
        sdl2.SDLK_END: uea.END, sdl2.SDLK_DOWN: uea.DOWN, sdl2.SDLK_PAGEDOWN: uea.PAGEDOWN,
        sdl2.SDLK_ESCAPE: uea.ESCAPE, sdl2.SDLK_BACKSPACE: uea.BACKSPACE, sdl2.SDLK_TAB: uea.TAB,
        sdl2.SDLK_RETURN: uea.RETURN, sdl2.SDLK_KP_ENTER: uea.RETURN, sdl2.SDLK_SPACE: uea.SPACE,
        sdl2.SDLK_INSERT: uea.INSERT, sdl2.SDLK_DELETE: uea.DELETE,
    }

    SHIFT_KEY_TO_EASCII = {
        sdl2.SDLK_F1: uea.SHIFT_F1, sdl2.SDLK_F2: uea.SHIFT_F2, sdl2.SDLK_F3: uea.SHIFT_F3,
        sdl2.SDLK_F4: uea.SHIFT_F4, sdl2.SDLK_F5: uea.SHIFT_F5, sdl2.SDLK_F6: uea.SHIFT_F6,
        sdl2.SDLK_F7: uea.SHIFT_F7, sdl2.SDLK_F8: uea.SHIFT_F8, sdl2.SDLK_F9: uea.SHIFT_F9,
        sdl2.SDLK_F10: uea.SHIFT_F10, sdl2.SDLK_F11: uea.SHIFT_F11, sdl2.SDLK_F12: uea.SHIFT_F12,
        sdl2.SDLK_HOME: uea.SHIFT_HOME, sdl2.SDLK_UP: uea.SHIFT_UP,
        sdl2.SDLK_PAGEUP: uea.SHIFT_PAGEUP, sdl2.SDLK_LEFT: uea.SHIFT_LEFT,
        sdl2.SDLK_RIGHT: uea.SHIFT_RIGHT, sdl2.SDLK_END: uea.SHIFT_END,
        sdl2.SDLK_DOWN: uea.SHIFT_DOWN, sdl2.SDLK_PAGEDOWN: uea.SHIFT_PAGEDOWN,
        sdl2.SDLK_ESCAPE: uea.SHIFT_ESCAPE, sdl2.SDLK_BACKSPACE: uea.SHIFT_BACKSPACE,
        sdl2.SDLK_TAB: uea.SHIFT_TAB, sdl2.SDLK_RETURN: uea.SHIFT_RETURN,
        sdl2.SDLK_KP_ENTER: uea.SHIFT_RETURN, sdl2.SDLK_SPACE: uea.SHIFT_SPACE,
        sdl2.SDLK_INSERT: uea.SHIFT_INSERT, sdl2.SDLK_DELETE: uea.SHIFT_DELETE,
        sdl2.SDLK_KP_5: uea.SHIFT_KP5,
    }

    CTRL_KEY_TO_EASCII = {
        sdl2.SDLK_F1: uea.CTRL_F1, sdl2.SDLK_F2: uea.CTRL_F2, sdl2.SDLK_F3: uea.CTRL_F3,
        sdl2.SDLK_F4: uea.CTRL_F4, sdl2.SDLK_F5: uea.CTRL_F5, sdl2.SDLK_F6: uea.CTRL_F6,
        sdl2.SDLK_F7: uea.CTRL_F7, sdl2.SDLK_F8: uea.CTRL_F8, sdl2.SDLK_F9: uea.CTRL_F9,
        sdl2.SDLK_F10: uea.CTRL_F10, sdl2.SDLK_F11: uea.CTRL_F11, sdl2.SDLK_F12: uea.CTRL_F12,
        sdl2.SDLK_HOME: uea.CTRL_HOME, sdl2.SDLK_PAGEUP: uea.CTRL_PAGEUP,
        sdl2.SDLK_LEFT: uea.CTRL_LEFT, sdl2.SDLK_RIGHT: uea.CTRL_RIGHT, sdl2.SDLK_END: uea.CTRL_END,
        sdl2.SDLK_PAGEDOWN: uea.CTRL_PAGEDOWN, sdl2.SDLK_ESCAPE: uea.CTRL_ESCAPE,
        sdl2.SDLK_BACKSPACE: uea.CTRL_BACKSPACE, sdl2.SDLK_TAB: uea.CTRL_TAB,
        sdl2.SDLK_RETURN: uea.CTRL_RETURN, sdl2.SDLK_KP_ENTER: uea.CTRL_RETURN,
        sdl2.SDLK_SPACE: uea.CTRL_SPACE, sdl2.SDLK_PRINTSCREEN: uea.CTRL_PRINT,
        sdl2.SDLK_2: uea.CTRL_2, sdl2.SDLK_6: uea.CTRL_6, sdl2.SDLK_MINUS: uea.CTRL_MINUS,
    }

    ALT_SCAN_TO_EASCII = {
        sdl2.SDL_SCANCODE_1: uea.ALT_1, sdl2.SDL_SCANCODE_2: uea.ALT_2,
        sdl2.SDL_SCANCODE_3: uea.ALT_3, sdl2.SDL_SCANCODE_4: uea.ALT_4,
        sdl2.SDL_SCANCODE_5: uea.ALT_5, sdl2.SDL_SCANCODE_6: uea.ALT_6,
        sdl2.SDL_SCANCODE_7: uea.ALT_7, sdl2.SDL_SCANCODE_8: uea.ALT_8,
        sdl2.SDL_SCANCODE_9: uea.ALT_9, sdl2.SDL_SCANCODE_0: uea.ALT_0,
        sdl2.SDL_SCANCODE_MINUS: uea.ALT_MINUS, sdl2.SDL_SCANCODE_EQUALS: uea.ALT_EQUALS,
        sdl2.SDL_SCANCODE_Q: uea.ALT_q, sdl2.SDL_SCANCODE_W: uea.ALT_w,
        sdl2.SDL_SCANCODE_E: uea.ALT_e, sdl2.SDL_SCANCODE_R: uea.ALT_r,
        sdl2.SDL_SCANCODE_T: uea.ALT_t, sdl2.SDL_SCANCODE_Y: uea.ALT_y,
        sdl2.SDL_SCANCODE_U: uea.ALT_u, sdl2.SDL_SCANCODE_I: uea.ALT_i,
        sdl2.SDL_SCANCODE_O: uea.ALT_o, sdl2.SDL_SCANCODE_P: uea.ALT_p,
        sdl2.SDL_SCANCODE_A: uea.ALT_a, sdl2.SDL_SCANCODE_S: uea.ALT_s,
        sdl2.SDL_SCANCODE_D: uea.ALT_d, sdl2.SDL_SCANCODE_F: uea.ALT_f,
        sdl2.SDL_SCANCODE_G: uea.ALT_g, sdl2.SDL_SCANCODE_H: uea.ALT_h,
        sdl2.SDL_SCANCODE_J: uea.ALT_j, sdl2.SDL_SCANCODE_K: uea.ALT_k,
        sdl2.SDL_SCANCODE_L: uea.ALT_l, sdl2.SDL_SCANCODE_Z: uea.ALT_z,
        sdl2.SDL_SCANCODE_X: uea.ALT_x, sdl2.SDL_SCANCODE_C: uea.ALT_c,
        sdl2.SDL_SCANCODE_V: uea.ALT_v, sdl2.SDL_SCANCODE_B: uea.ALT_b,
        sdl2.SDL_SCANCODE_N: uea.ALT_n, sdl2.SDL_SCANCODE_M: uea.ALT_m,
        sdl2.SDL_SCANCODE_F1: uea.ALT_F1, sdl2.SDL_SCANCODE_F2: uea.ALT_F2,
        sdl2.SDL_SCANCODE_F3: uea.ALT_F3, sdl2.SDL_SCANCODE_F4: uea.ALT_F4,
        sdl2.SDL_SCANCODE_F5: uea.ALT_F5, sdl2.SDL_SCANCODE_F6: uea.ALT_F6,
        sdl2.SDL_SCANCODE_F7: uea.ALT_F7, sdl2.SDL_SCANCODE_F8: uea.ALT_F8,
        sdl2.SDL_SCANCODE_F9: uea.ALT_F9, sdl2.SDL_SCANCODE_F10: uea.ALT_F10,
        sdl2.SDL_SCANCODE_F11: uea.ALT_F11, sdl2.SDL_SCANCODE_F12: uea.ALT_F12,
        sdl2.SDL_SCANCODE_BACKSPACE: uea.ALT_BACKSPACE, sdl2.SDL_SCANCODE_TAB: uea.ALT_TAB,
        sdl2.SDL_SCANCODE_RETURN: uea.ALT_RETURN, sdl2.SDL_SCANCODE_KP_ENTER: uea.ALT_RETURN,
        sdl2.SDL_SCANCODE_SPACE: uea.ALT_SPACE, sdl2.SDL_SCANCODE_PRINTSCREEN: uea.ALT_PRINT,
        sdl2.SDL_SCANCODE_KP_5: uea.ALT_KP5,
    }

    MOD_TO_SCAN = {
        sdl2.KMOD_LSHIFT: scancode.LSHIFT, sdl2.KMOD_RSHIFT: scancode.RSHIFT,
        sdl2.KMOD_LCTRL: scancode.CTRL, sdl2.KMOD_RCTRL: scancode.CTRL,
        sdl2.KMOD_LALT: scancode.ALT,
        # don't catch right-Alt as it may inhibit AltGr on Windows
        #sdl2.KMOD_RALT: scancode.ALT,
    }


###############################################################################
# clipboard handling


class SDL2Clipboard(clipboard.Clipboard):
    """Clipboard handling interface using SDL2."""

    def __init__(self):
        """Initialise the clipboard handler."""
        clipboard.Clipboard.__init__(self)
        self.ok = (sdl2 is not None)

    def copy(self, text):
        """Put unicode text on clipboard."""
        sdl2.SDL_SetClipboardText(text.encode('utf-8', errors='replace'))

    def paste(self):
        """Return unicode text from clipboard."""
        text = sdl2.SDL_GetClipboardText()
        if text is None:
            return u''
        return text.decode('utf-8', 'replace').replace(u'\r\n', u'\n').replace(u'\n', u'\r')


###############################################################################

def _pixels2d(surface_ptr):
    """Creates a 2D pixel array view of the passed 8-bit surface."""
    # based on pysdl2.ext.pixels2d by Marcus von Appen
    # original is CC0 public domain with zlib fallback licence
    # https://bitbucket.org/marcusva/py-sdl2
    surface = surface_ptr.contents
    srcsize = surface.h * surface.pitch
    pxbuf = ctypes.cast(surface.pixels, ctypes.POINTER(ctypes.c_ubyte * srcsize)).contents
    # I don't understand this line, but I need it in Python 3
    # see https://bugs.python.org/issue15944
    if not PY2:
        pxbuf = memoryview(pxbuf).cast('B')
    return bytematrix.ByteMatrix.view_from_buffer(surface.h, surface.w, surface.pitch, pxbuf)


###############################################################################
# video plugin

@video_plugins.register('sdl2')
class VideoSDL2(VideoPlugin):
    """SDL2-based graphical interface."""

    def __init__(
            self, input_queue, video_queue,
            caption=u'', icon=BLANK_ICON,
            scaling=None, dimensions=None, aspect_ratio=(4, 3), border_width=0, fullscreen=False,
            prevent_close=False, mouse_clipboard=True,
            **kwargs
        ):
        """Initialise SDL2 interface."""
        try:
            _import_sdl2()
        except ImportError:
            raise InitFailed('Module `sdl2` not found')
        VideoPlugin.__init__(self, input_queue, video_queue)
        # Windows 10 - set to DPI aware to avoid scaling twice on HiDPI screens
        set_dpi_aware()
        # request smooth scaling
        self._smooth = scaling == 'smooth'
        # ignore ALT+F4 and window X button
        self._nokill = prevent_close
        # window caption/title
        self._caption = caption
        # start in fullscreen mode if True
        self._fullscreen = fullscreen
        # don't resize native windows
        self._resizable = scaling != 'native'
        # display & border
        # border attribute
        self._border_attr = 0
        # update cycle
        self._cycle = 0
        self._last_tick = 0
        # blink is enabled, should be True in text modes with blink and ega mono
        # set to true if blinking attributes occur in palette
        self._palette_blinks = False
        # load the icon
        self._icon = bytematrix.ByteMatrix(len(icon), len(icon[0]), icon).hrepeat(2).vrepeat(2)
        # mouse setups
        self._mouse_clip = mouse_clipboard
        # keyboard setup
        self._f11_active = False
        self._env = EnvironmentCache()
        if WIN32:
            # ensure the correct SDL2 video driver is chosen for Windows
            # since this gets messed up if we also import pygame
            self._env.set('SDL_VIDEODRIVER', 'windows')
        else:
            # this is needed "to make the window manager correctly match the sdl window" (see pygame source)
            # without it, the icon doesn't get changed
            self._env.set('SDL_VIDEO_X11_WMCLASS', sys.argv[0])
        # initialise SDL
        if sdl2.SDL_Init(sdl2.SDL_INIT_EVERYTHING):
            # SDL not initialised correctly
            # reset the environment variable
            # to not throw PyGame off if we try that later
            self._env.close()
            raise InitFailed('Could not initialise SDL2: %s' % sdl2.SDL_GetError())
        # get physical screen dimensions (needs to be called before set_mode)
        display_mode = sdl2.SDL_DisplayMode()
        sdl2.SDL_GetCurrentDisplayMode(0, ctypes.byref(display_mode))
        self._window_sizer = window.WindowSizer(
            display_mode.w, display_mode.h,
            scaling, dimensions, aspect_ratio, border_width,
        )
        # create the window initially as 720*400 black
        self._window_sizer.set_canvas_size(720, 400, fullscreen=self._fullscreen)
        # canvas surfaces
        self._window_surface = None
        # pixel views of canvases
        self._canvas_pixels = None
        # main window object
        self._display = None
        self._display_surface = None
        # has mode been set at least once?
        self._mode_set = False
        # one cache per blink state
        self._display_cache = [None] * N_BLINK_STATES
        self._has_display_cache = [False] * N_BLINK_STATES
        # pointer to the zoomed surface
        self._zoomed_surface = None
        # clipboard handler
        self._clipboard_handler = None
        # clipboard visual feedback
        self._clipboard_interface = None
        # event handlers
        self._event_handlers = self._register_handlers()
        # video mode settings
        self._text_cursor = True
        self._font_height = None
        self._font_width = None
        # cursor
        # cursor is visible
        self._cursor_visible = True
        # cursor position
        self._cursor_row, self._cursor_col = 1, 1
        # cursor shape
        self._cursor_from = None
        self._cursor_height = None
        self._cursor_width = None
        self._cursor_attr = None
        self._cursor_cache = [None, None]
        # palette
        # { attr: (fore, back, blink, underline) }
        self._attributes = {}
        # display palettes for blink states 0, 1
        self._palette = [sdl2.SDL_AllocPalette(256), sdl2.SDL_AllocPalette(256)]
        # pixel packing is active (composite artifacts)
        self._pixel_packing = False
        # last keypress
        self._last_keypress = None
        # set clipboard handler to SDL2
        self._clipboard_handler = SDL2Clipboard()
        # available joysticks
        num_joysticks = sdl2.SDL_NumJoysticks()
        for stick in range(num_joysticks):
            sdl2.SDL_JoystickOpen(stick)
            # if a joystick is present, its axes report 128 for mid, not 0
            for axis in (0, 1):
                self._input_queue.put(signals.Event(signals.STICK_MOVED, (stick, axis, 128)))

    def __enter__(self):
        """Complete SDL2 interface initialisation."""
        # "NOTE: You should not expect to be able to create a window, render,
        #        or receive events on any thread other than the main one"
        # https://wiki.libsdl.org/CategoryThread
        # http://stackoverflow.com/questions/27751533/sdl2-threading-seg-fault
        self._do_create_window()
        # pop up as black rather than background, looks nicer
        sdl2.SDL_UpdateWindowSurface(self._display)
        # check if we can honour scaling=smooth
        if self._smooth:
            bpp = self._display_surface.contents.format.contents.BitsPerPixel
            if bpp != 32:
                logging.warning(
                    'Smooth scaling not available: need 32-bit colour, have %d-bit.', bpp
                )
                self._smooth = False
            if not hasattr(sdl2, 'sdlgfx') or not sdl2.sdlgfx.zoomSurface:
                logging.warning('Smooth scaling not available: `sdlgfx` extension not found.')
                self._smooth = False
        # enable IME
        sdl2.SDL_StartTextInput()
        return VideoPlugin.__enter__(self)
        # set_mode should be first event on queue

    def __exit__(self, exc_type, value, traceback):
        """Close the SDL2 interface."""
        VideoPlugin.__exit__(self, exc_type, value, traceback)
        if not sdl2:
            return
        if self._display:
            # free windows
            sdl2.SDL_DestroyWindow(self._display)
        # free caches
        for surface in self._display_cache:
            sdl2.SDL_FreeSurface(surface)
        # free surfaces
        if self._window_surface:
            sdl2.SDL_FreeSurface(self._window_surface)
        # free palettes
        for palette in self._palette:
            sdl2.SDL_FreePalette(palette)
        # close IME
        sdl2.SDL_StopTextInput()
        # close SDL2
        sdl2.SDL_Quit()

    def _set_icon(self):
        """Set the icon on the SDL window."""
        icon = sdl2.SDL_CreateRGBSurface(0, self._icon.width, self._icon.height, 8, 0, 0, 0, 0)
        _pixels2d(icon)[:, :] = self._icon
        # icon palette (black & white)
        icon_palette = sdl2.SDL_AllocPalette(256)
        icon_colors = [sdl2.SDL_Color(_c, _c, _c, 255) for _c in [0, 255] + [255]*254]
        sdl2.SDL_SetPaletteColors(icon_palette, (sdl2.SDL_Color * 256)(*icon_colors), 0, 2)
        sdl2.SDL_SetSurfacePalette(icon, icon_palette)
        sdl2.SDL_SetWindowIcon(self._display, icon)
        sdl2.SDL_FreeSurface(icon)
        sdl2.SDL_FreePalette(icon_palette)

    def _do_create_window(self):
        """Create a new SDL window """
        flags = sdl2.SDL_WINDOW_SHOWN
        if self._resizable:
            flags |= sdl2.SDL_WINDOW_RESIZABLE
        if self._fullscreen:
            flags |= sdl2.SDL_WINDOW_FULLSCREEN_DESKTOP | sdl2.SDL_WINDOW_BORDERLESS
        width, height = self._window_sizer.display_size
        sdl2.SDL_DestroyWindow(self._display)
        self._display = sdl2.SDL_CreateWindow(
            self._caption.encode('utf-8', errors='replace'),
            sdl2.SDL_WINDOWPOS_CENTERED, sdl2.SDL_WINDOWPOS_CENTERED,
            width, height, flags
        )
        # on fullscreen, grab keyboard exclusively
        # this allows BASIC to capture Alt-F4, Alt-TAB etc.
        if self._fullscreen:
            sdl2.SDL_SetHint(sdl2.SDL_HINT_GRAB_KEYBOARD, b'1')
            sdl2.SDL_SetWindowGrab(self._display, sdl2.SDL_TRUE)
        self._set_icon()
        self._reset_display_caches()
        self.busy = True

    def _reset_display_caches(self):
        """Reset caches and references to display object."""
        # this is needed on macOS, we get a null pointer result if not set
        # and SDL_GetError gives a confusing "No hardware accelerated renderers available"
        sdl2.SDL_SetHint(b'SDL_FRAMEBUFFER_ACCELERATION', b'opengl')
        self._display_surface = sdl2.SDL_GetWindowSurface(self._display)
        # reset cache sizes
        for surface in self._display_cache:
            sdl2.SDL_FreeSurface(surface)
        # clone the surface four times for blink caches
        self._display_cache = [
            sdl2.SDL_ConvertSurface(self._display_surface, self._display_surface.contents.format, 0)
            for _ in range(N_BLINK_STATES)
        ]
        self._has_display_cache = [False] * N_BLINK_STATES


    ###########################################################################
    # input cycle

    def _check_input(self):
        """Handle screen and interface events."""
        # don't try to handle events before display is set up
        if not self._display or not self._mode_set:
            return
        # check and handle input events
        self._last_keypress = None
        event = sdl2.SDL_Event()
        while sdl2.SDL_PollEvent(ctypes.byref(event)):
            try:
                self._event_handlers[event.type](event)
            except KeyError:
                pass
        self._flush_keypress()

    def _register_handlers(self):
        """Create table of event handlers."""
        return {
            sdl2.SDL_KEYDOWN: self._handle_key_down,
            sdl2.SDL_KEYUP: self._handle_key_up,
            sdl2.SDL_TEXTINPUT: self._handle_text_input,
            sdl2.SDL_TEXTEDITING: self._handle_text_editing,
            sdl2.SDL_MOUSEBUTTONDOWN: self._handle_mouse_down,
            sdl2.SDL_MOUSEBUTTONUP: self._handle_mouse_up,
            sdl2.SDL_MOUSEMOTION: self._handle_mouse_motion,
            sdl2.SDL_JOYBUTTONDOWN: self._handle_stick_down,
            sdl2.SDL_JOYBUTTONUP: self._handle_stick_up,
            sdl2.SDL_JOYAXISMOTION: self._handle_stick_motion,
            sdl2.SDL_WINDOWEVENT: self._handle_window_event,
            sdl2.SDL_QUIT: self._handle_quit,
        }

    # quit events

    def _handle_quit(self, event):
        """Handle quit event."""
        if self._nokill:
            self.set_caption_message(NOKILL_MESSAGE)
        else:
            self._input_queue.put(signals.Event(signals.QUIT))

    # window events

    def _handle_window_event(self, event):
        """Handle window event."""
        if event.window.event == sdl2.SDL_WINDOWEVENT_RESIZED:
            self._handle_resize_event(event)
        # unset Alt modifiers on entering/leaving the window
        # workaround for what seems to be an SDL2 bug
        # where the ALT modifier sticks on the first Alt-Tab out
        # of the window
        elif event.window.event in (
                sdl2.SDL_WINDOWEVENT_LEAVE,
                sdl2.SDL_WINDOWEVENT_ENTER,
                sdl2.SDL_WINDOWEVENT_FOCUS_LOST,
                sdl2.SDL_WINDOWEVENT_FOCUS_GAINED
            ):
            sdl2.SDL_SetModState(sdl2.SDL_GetModState() & ~sdl2.KMOD_ALT)

    def _handle_resize_event(self, event):
        """Respond to change of display size."""
        if not self._fullscreen:
            # width, height = event.window.data1, event.window.data2
            # get actual window size
            width, height = ctypes.c_int(), ctypes.c_int()
            sdl2.SDL_GetWindowSize(self._display, ctypes.byref(width), ctypes.byref(height))
            # update the size calculator
            self._window_sizer.set_display_size(width.value, height.value)
        # we need to update the surface pointer
        self._reset_display_caches()
        self.busy = True

    # mouse events

    def _handle_mouse_down(self, event):
        """Handle mouse-down event."""
        pos = self._window_sizer.normalise_pos(event.button.x, event.button.y)
        if self._mouse_clip:
            if event.button.button == sdl2.SDL_BUTTON_LEFT:
                # LEFT button: copy
                self._clipboard_interface.start(
                    1 + pos[1] // self._font_height,
                    1 + (pos[0]+self._font_width//2) // self._font_width
                )
            elif event.button.button == sdl2.SDL_BUTTON_MIDDLE:
                # MIDDLE button: paste
                text = self._clipboard_handler.paste()
                self._clipboard_interface.paste(text)
            self.busy = True
        if event.button.button == sdl2.SDL_BUTTON_LEFT:
            # pen press
            self._input_queue.put(signals.Event(signals.PEN_DOWN, pos))

    def _handle_mouse_up(self, event):
        """Handle mouse-up event."""
        self._input_queue.put(signals.Event(signals.PEN_UP))
        if self._mouse_clip and event.button.button == sdl2.SDL_BUTTON_LEFT:
            self._clipboard_interface.copy()
            self._clipboard_interface.stop()
            self.busy = True

    def _handle_mouse_motion(self, event):
        """Handle mouse-motion event."""
        pos = self._window_sizer.normalise_pos(event.motion.x, event.motion.y)
        self._input_queue.put(signals.Event(signals.PEN_MOVED, pos))
        if self._clipboard_interface.active():
            self._clipboard_interface.move(
                1 + pos[1] // self._font_height,
                1 + (pos[0]+self._font_width//2) // self._font_width
            )
            self.busy = True

    # joystick events

    def _handle_stick_down(self, event):
        """Handle joystick button-down event."""
        self._input_queue.put(signals.Event(
            signals.STICK_DOWN,
            (event.jbutton.which, event.jbutton.button)
        ))

    def _handle_stick_up(self, event):
        """Handle joystick button-up event."""
        self._input_queue.put(signals.Event(
            signals.STICK_UP,
            (event.jbutton.which, event.jbutton.button)
        ))

    def _handle_stick_motion(self, event):
        """Handle joystick axis-motion event."""
        self._input_queue.put(signals.Event(
            signals.STICK_MOVED,
            (event.jaxis.which, event.jaxis.axis, int((event.jaxis.value/32768.)*127 + 128))
        ))

    # keyboard events

    def _handle_key_down(self, event):
        """Handle key-down event."""
        # get scancode
        scan = SCAN_TO_SCAN.get(event.key.keysym.scancode, None)
        # get modifiers
        mod = [_s for _m, _s in iteritems(MOD_TO_SCAN) if event.key.keysym.mod & _m]
        # get eascii
        try:
            if scancode.ALT in mod:
                char = ALT_SCAN_TO_EASCII[event.key.keysym.scancode]
            elif scancode.CTRL in mod:
                char = CTRL_KEY_TO_EASCII[event.key.keysym.sym]
            elif scancode.LSHIFT in mod or scancode.RSHIFT in mod:
                char = SHIFT_KEY_TO_EASCII[event.key.keysym.sym]
            else:
                char = KEY_TO_EASCII[event.key.keysym.sym]
        except KeyError:
            # try control+letter -> control codes
            key = event.key.keysym.sym
            if scancode.CTRL in mod and key >= ord(u'a') and key <= ord(u'z'):
                char = unichr(key - ord(u'a') + 1)
            elif scancode.CTRL in mod and key >= ord(u'[') and key <= ord(u'_'):
                char = unichr(key - ord(u'A') + 1)
            else:
                char = u''
        if char == u'\0':
            char = uea.NUL
        # handle F11 home-key combinations
        if event.key.keysym.sym == sdl2.SDLK_F11:
            self._f11_active = True
            self._clipboard_interface.start(self._cursor_row, self._cursor_col)
        elif self._f11_active:
            self._clipboard_interface.handle_key(scan, char)
            self.busy = True
        else:
            # keep scancode in last-down buffer to combine with text event
            # flush buffer on next key down, text event or end of loop
            # if the same key is reported twice with the same timestamp, ignore
            # (this deals with the Unity double-ALT-bug and maybe others)
            if scan is not None and (char, scan, mod, event.key.timestamp) != self._last_keypress:
                self._flush_keypress()
                self._last_keypress = char, scan, mod, event.key.timestamp

    def _flush_keypress(self):
        """Flush last keypress from buffer."""
        if self._last_keypress is not None:
            # insert into keyboard queue; no text event
            char, scan, mod, _ = self._last_keypress
            self._input_queue.put(signals.Event(signals.KEYB_DOWN, (char, scan, mod)))
            self._last_keypress = None

    def _handle_key_up(self, event):
        """Handle key-up event."""
        try:
            scan = SCAN_TO_SCAN[event.key.keysym.scancode]
        except KeyError:
            return
        # check for emulator key
        if event.key.keysym.sym == sdl2.SDLK_F11:
            self._clipboard_interface.stop()
            self.busy = True
            self._f11_active = False
        # last key released gets remembered
        try:
            self._input_queue.put(signals.Event(signals.KEYB_UP, (scan,)))
        except KeyError:
            pass

    # text input method events

    def _handle_text_editing(self, event):
        """Handle text-editing event."""
        self.set_caption_message(event.text.text.decode('utf-8'))

    def _handle_text_input(self, event):
        """Handle text-input event."""
        char = event.text.text.decode('utf-8', errors='replace')
        if self._f11_active:
            # F11+f to toggle fullscreen mode
            if char.upper() == u'F':
                self._toggle_fullscreen()
            self._clipboard_interface.handle_key(None, char)
        # the text input event follows the key down event immediately
        elif self._last_keypress is None:
            # no key down event waiting: other input method
            self._input_queue.put(signals.Event(signals.KEYB_DOWN, (char, None, None)))
        else:
            eascii, scan, mod, timestamp = self._last_keypress
            # timestamps for kepdown and textinput may differ by one on mac
            if timestamp + 1 >= event.text.timestamp:
                # combine if same time stamp
                if eascii and char != eascii:
                    # filter out chars being sent with alt+key on Linux
                    if scancode.ALT not in mod:
                        # with IME, the text is sent together with the final Enter keypress.
                        self._input_queue.put(signals.Event(signals.KEYB_DOWN, (char, None, None)))
                    else:
                        # final keypress such as space, CR have IME meaning, we should ignore them
                        self._input_queue.put(signals.Event(signals.KEYB_DOWN, (eascii, scan, mod)))
                else:
                    self._input_queue.put(signals.Event(signals.KEYB_DOWN, (char, scan, mod)))
            else:
                # two separate events
                # previous keypress has no corresponding textinput
                self._flush_keypress()
                # current textinput has no corresponding keypress
                self._input_queue.put(signals.Event(signals.KEYB_DOWN, (char, None, None)))
            self._last_keypress = None

    def _toggle_fullscreen(self):
        """Togggle fullscreen mode."""
        self._fullscreen = not self._fullscreen
        self._window_sizer.set_canvas_size(fullscreen=self._fullscreen)
        self._do_create_window()
        self.busy = True


    ###########################################################################
    # screen drawing cycle

    def sleep(self, ms):
        """Sleep a tick to avoid hogging the cpu."""
        sdl2.SDL_Delay(ms)

    def _work(self):
        """Check screen and blink events; update screen if necessary."""
        if not self._window_surface:
            return
        #               0      0      1      1
        # cycle         0 1234 5 6789 0 1234 5 6789 (0)
        # blink state   0 ---- 1 ---- 2 ---- 3 ---- (0)
        # cursor        off    on     off    on
        # blink         on     on     off    off
        #
        # blink state remains constant if blink not enabled
        # cursor blinks only if _text_cursor
        # cursor visible every cycle between 5 and 10, 15 and 20
        tick = sdl2.SDL_GetTicks()
        if tick - self._last_tick >= CYCLE_TIME:
            self._last_tick = tick
            self._cycle += 1
            if self._cycle == BLINK_CYCLES * N_BLINK_STATES:
                self._cycle = 0
            # blink state
            blink_state, blink_tock = divmod(self._cycle, BLINK_CYCLES)
            if not self._palette_blinks and not self._text_cursor:
                blink_state = 1
            # flip display fully if changed, use cache if just blinking
            if self.busy:
                self._clear_display_cache()
                self._flip_busy(blink_state)
                self.busy = False
            elif (self._palette_blinks or self._text_cursor) and blink_tock == 0:
                self._flip_lazy(blink_state)

    def _clear_display_cache(self):
        """Clear cursor cache on busy flip."""
        # one cache per blink state
        self._has_display_cache = [False] * N_BLINK_STATES

    def _flip_lazy(self, blink_state):
        """Blink the cursor only, to avoid doing all the scaling and converting work."""
        if self._has_display_cache[blink_state]:
            sdl2.SDL_BlitSurface(
                self._display_cache[blink_state], None, self._display_surface, None
            )
            sdl2.SDL_UpdateWindowSurface(self._display)
        else:
            # if we don't have a cache for this state, build it
            self._flip_busy(blink_state)

    def _flip_busy(self, blink_state):
        """Draw the canvas to the screen."""
        if self._pixel_packing:
            work_surface = self._create_composite_surface()
        else:
            work_surface = self._window_surface
        pixelformat = self._display_surface.contents.format
        # apply cursor to work surface
        with self._show_cursor((blink_state % 2) or not self._text_cursor):
            # convert 8-bit work surface to (usually) 32-bit display surface format
            sdl2.SDL_SetSurfacePalette(work_surface, self._palette[blink_state // 2])
            conv = sdl2.SDL_ConvertSurface(work_surface, pixelformat, 0)
        if self._pixel_packing:
            sdl2.SDL_FreeSurface(work_surface)
        # create clipboard feedback
        if self._clipboard_interface.active():
            self._show_clipboard(conv)
        # scale surface to final dimensions and flip
        self._scale_and_flip(conv, blink_state)
        # destroy the temporary surface
        sdl2.SDL_FreeSurface(conv)

    def _scale_and_flip(self, conv, blink_state):
        """Scale converted surface and flip onto display."""
        # determine letterbox dimensions
        xshift, yshift = self._window_sizer.letterbox_shift
        window_w, window_h = self._window_sizer.window_size
        target_rect = sdl2.SDL_Rect(xshift, yshift, window_w, window_h)
        if not self._smooth:
            sdl2.SDL_BlitScaled(conv, None, self._display_surface, target_rect)
        else:
            # smooth-scale converted surface
            scalex, scaley = self._window_sizer.scale
            # only free the surface just before zoomSurface needs to re-allocate
            # so that the memory block is highly likely to be easily available
            # this seems to avoid unpredictable delays
            sdl2.SDL_FreeSurface(self._zoomed_surface)
            # SMOOTHING_ON = 1
            self._zoomed_surface = sdl2.sdlgfx.zoomSurface(conv, scalex, scaley, 1)
            # blit onto display
            sdl2.SDL_BlitSurface(self._zoomed_surface, None, self._display_surface, target_rect)
        # save in display cache for this blink state
        sdl2.SDL_BlitSurface(self._display_surface, None, self._display_cache[blink_state], None)
        self._has_display_cache[blink_state] = True
        # flip the display
        sdl2.SDL_UpdateWindowSurface(self._display)

    @contextmanager
    def _show_cursor(self, cursor_state):
        """Draw or remove the cursor on the visible page."""
        if not self._cursor_visible or not cursor_state:
            yield
        else:
            # cursor shape
            top = (self._cursor_row-1) * self._font_height + self._cursor_from
            left = (self._cursor_col-1) * self._font_width
            cursor_slice = slice(top, top+self._cursor_height), slice(left, left+self._cursor_width)
            # canvas_pixels is a view
            cursor_area = self._canvas_pixels[cursor_slice]
            # copy area under cursor
            under_cursor = cursor_area.copy()
            if self._text_cursor:
                cursor_area[:, :] = self._cursor_attr
            else:
                cursor_area[:, :] ^= self._cursor_attr
            yield
            cursor_area[:, :] = under_cursor

    def _show_clipboard(self, conv):
        """Show clipboard feedback overlay."""
        n_rects = len(self._clipboard_interface.selection_rect)
        if not n_rects:
            return
        border_x, border_y = self._window_sizer.border_shift
        lcanvas_w, lcanvas_h = self._window_sizer.canvas_size_logical
        # create overlay for clipboard selection feedback
        overlay = sdl2.SDL_CreateRGBSurface(0, lcanvas_w, lcanvas_h, 32, 0, 0, 0, 0)
        sdl2.SDL_SetSurfaceBlendMode(overlay, sdl2.SDL_BLENDMODE_ADD)
        overlay_target = sdl2.SDL_Rect(border_x, border_y, lcanvas_w, lcanvas_h)
        rects = (sdl2.SDL_Rect * n_rects)(*(
            sdl2.SDL_Rect(*r) for r in self._clipboard_interface.selection_rect
        ))
        sdl2.SDL_FillRects(
            overlay, rects, n_rects,
            sdl2.SDL_MapRGBA(overlay.contents.format, 128, 0, 128, 0)
        )
        sdl2.SDL_BlitSurface(overlay, None, conv, overlay_target)
        sdl2.SDL_FreeSurface(overlay)

    def _create_composite_surface(self):
        """Pack multiple pixels into one for composite artifacts."""
        lwindow_w, lwindow_h = self._window_sizer.window_size_logical
        border_x, border_y = self._window_sizer.border_shift
        # pack pixels into higher bpp, then unpack into lower bpp
        bpp_out, bpp_in = self._pixel_packing
        packed = self._canvas_pixels.packed(8//bpp_in)
        height = lwindow_h - border_y*2
        unpacked = bytematrix.ByteMatrix.frompacked(packed, height, 8//bpp_out)
        unpacked = unpacked.hrepeat(bpp_out // bpp_in)
        # copy packed array onto work surface
        work_surface = sdl2.SDL_CreateRGBSurface(0, lwindow_w, lwindow_h, 8, 0, 0, 0, 0)
        _pixels2d(work_surface)[
            border_y : lwindow_h - border_y,
            border_x : lwindow_w - border_x
        ] = unpacked
        return work_surface


    ###########################################################################
    # signal handlers

    def set_mode(self, canvas_height, canvas_width, text_height, text_width):
        """Initialise a given text or graphics mode."""
        # set geometry
        self._font_height = -(-canvas_height // text_height)
        self._font_width = canvas_width // text_width
        # set standard cursor
        self._cursor_width = self._font_width
        self._cursor_from, self._cursor_height = 0, self._font_height
        # logical size
        size_changed = self._window_sizer.set_canvas_size(
            canvas_width, canvas_height, fullscreen=self._fullscreen, resize_window=False
        )
        # only ever adjust window size if we're in native pixel mode
        if size_changed:
            if self._fullscreen:
                # clear any areas now outside the window
                sdl2.SDL_FillRect(self._display_surface, None, 0)
            else:
                # resize and recentre
                sdl2.SDL_SetWindowSize(self._display, *self._window_sizer.display_size)
                sdl2.SDL_SetWindowPosition(
                    self._display, sdl2.SDL_WINDOWPOS_CENTERED, sdl2.SDL_WINDOWPOS_CENTERED
                )
                # need to update surface pointer after a change in window size
                self._reset_display_caches()
        # screen pages
        if self._window_surface:
            sdl2.SDL_FreeSurface(self._window_surface)
        work_width, work_height = self._window_sizer.window_size_logical
        self._window_surface = sdl2.SDL_CreateRGBSurface(0, work_width, work_height, 8, 0, 0, 0, 0)
        border_x, border_y = self._window_sizer.border_shift
        self._canvas_pixels = _pixels2d(self._window_surface)[
            border_y : work_height - border_y, border_x : work_width - border_x
        ]
        # initialise clipboard
        self._clipboard_interface = clipboard.ClipboardInterface(
            self._clipboard_handler, self._input_queue,
            text_width, text_height, self._font_width, self._font_height,
            (canvas_width, canvas_height)
        )
        self._mode_set = True
        self.busy = True

    def set_caption_message(self, msg):
        """Add a message to the window caption."""
        title = self._caption + (u' - ' + msg if msg else u'')
        sdl2.SDL_SetWindowTitle(self._display, title.encode('utf-8', errors='replace'))

    def set_clipboard_text(self, text):
        """Put text on the clipboard."""
        self._clipboard_handler.copy(text)

    def set_palette(self, attributes, pack_pixels):
        """Build the palette."""
        self._attributes = attributes
        palette_blink_up = [_fore for _fore, _, _, _ in attributes]
        palette_blink_down = [_back if _blink else _fore for _fore, _back, _blink, _ in attributes]
        self._palette_blinks = palette_blink_up != palette_blink_down
        # blink states: 0 light up, 1 light down
        colors_0 = (sdl2.SDL_Color * 256)(*(
            sdl2.SDL_Color(_r, _g, _b, 255)
            for (_r, _g, _b) in palette_blink_up
        ))
        colors_1 = (sdl2.SDL_Color * 256)(*(
            sdl2.SDL_Color(_r, _g, _b, 255)
            for (_r, _g, _b) in palette_blink_down
        ))
        sdl2.SDL_SetPaletteColors(self._palette[0], colors_0, 0, 256)
        sdl2.SDL_SetPaletteColors(self._palette[1], colors_1, 0, 256)
        self._pixel_packing = pack_pixels
        self.busy = True

    def set_border_attr(self, attr):
        """Change the border attribute."""
        window_w, window_h = self._window_sizer.window_size_logical
        border_x, border_y = self._window_sizer.border_shift
        border_rects = (sdl2.SDL_Rect*4)(
            sdl2.SDL_Rect(0, 0, window_w, border_y),
            sdl2.SDL_Rect(0, 0, border_x, window_h),
            sdl2.SDL_Rect(window_w-border_x, 0, border_x, window_h),
            sdl2.SDL_Rect(0, window_h-border_y, window_w, border_y),
        )
        if self._window_surface:
            sdl2.SDL_FillRects(self._window_surface, border_rects, 4, attr)
        self._border_attr = attr
        self.busy = True

    def clear_rows(self, back_attr, start, stop):
        """Clear a range of screen rows."""
        self._canvas_pixels[
            (start-1)*self._font_height : stop*self._font_height,
            0 : self._window_sizer.width
        ] = back_attr
        self.busy = True

    def show_cursor(self, cursor_on, cursor_blinks):
        """Change visibility of cursor."""
        self._text_cursor = cursor_blinks
        self._cursor_visible = cursor_on
        self.busy = True

    def move_cursor(self, row, col, attr, width):
        """Move the cursor to a new position."""
        if self._cursor_visible and (
                self._cursor_row, self._cursor_col, self._cursor_attr, self._cursor_width
            ) != (row, col, attr, width):
            self.busy = True
        self._cursor_row, self._cursor_col = row, col
        self._cursor_attr = attr
        self._cursor_width = width

    def set_cursor_shape(self, from_line, to_line):
        """Build a sprite for the cursor."""
        self._cursor_from = from_line
        self._cursor_height = to_line + 1 - from_line
        if self._cursor_visible:
            self.busy = True

    def scroll(self, direction, from_line, scroll_height, back_attr):
        """Scroll the screen between from_line and scroll_height."""
        pixels = self._canvas_pixels
        # scroll window, top of rows
        hi_y0, hi_y1 = (from_line-1)*self._font_height, (scroll_height-1)*self._font_height
        # scroll window, bottom of rows
        lo_y0, lo_y1 = from_line*self._font_height, scroll_height*self._font_height
        if direction == -1:
            # scroll up
            pixels[hi_y0:hi_y1, :] = pixels[lo_y0:lo_y1, :]
            # clear the new empty line
            pixels[hi_y1:lo_y1, :] = back_attr
        else:
            # scroll down
            # copy is needed here as bytearray-view self-slice assignment will self-overwrite
            pixels[lo_y0:lo_y1, :] = pixels[hi_y0:hi_y1, :].copy()
            # clear the new empty line
            pixels[hi_y0:lo_y0, :] = back_attr
        self.busy = True

    def update(self, row, col, unicode_matrix, attr_matrix, y0, x0, sprite):
        """Put text or pixels at a given position."""
        if not sprite:
            return
        # reference the destination area
        pixels = self._canvas_pixels
        # clip to size if needed
        if y0 + sprite.height > pixels.height or x0 + sprite.width > pixels.width:
            sprite = sprite[:pixels.height-y0, :pixels.width-x0]
        pixels[y0:y0+sprite.height, x0:x0+sprite.width] = sprite
        self.busy = True



"""
PC-BASIC - guard
Crash guard

(c) 2013--2023 Rob Hagemans
This file is released under the GNU GPL version 3 or later.
"""

import os
import io
import sys
import logging
import platform
import tempfile
import traceback
import webbrowser
import json
from datetime import datetime
from contextlib import contextmanager
from subprocess import check_output, CalledProcessError

from .basic.base import error, signals
from .basic import VERSION, LONG_VERSION, Session
from .compat import BrokenPipeError, is_broken_pipe, text_type
from . import info


class ExceptionGuard(object):
    """Context manager to handle uncaught exceptions."""

    def __init__(self, interface, log_dir=u'', uargv=()):
        """Initialise crash guard."""
        self._interface = interface
        self._uargv = uargv
        self._log_dir = log_dir
        self._session = None

    def __call__(self, session):
        """Complete initialisation."""
        self._session = session
        return self

    def __enter__(self):
        """Enter context guard."""
        self.exception_handled = None
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        """Handle exceptions."""
        success = False
        if not exc_type or exc_type == error.Reset:
            return success
        if is_broken_pipe(exc_value):
            # BrokenPipeError may be raised by shell pipes, handled at entry point
            # see docs.python.org/3/library/signal.html#note-on-sigpipe
            return success
        try:
            success = _bluescreen(
                self._session, self._interface,
                self._uargv, self._log_dir,
                exc_type, exc_value, exc_traceback
            )
            if success:
                # we still want to show the exception on the log
                logging.error(''.join(traceback.format_exception(exc_type, exc_value, exc_traceback)))
                self.exception_handled = exc_value
        except error.Exit:
            pass
        except BaseException as e:
            logging.error(e)
            raise
        return success


LOG_PATTERN = u'crash-%Y%m%d-'
CAPTION = u'System error. Please file a bug report. Press <Enter> to resume.'

REPORT_TEMPLATE="""
10 ' ** modal crash report **
20 SCREEN 0,0,0,0: WIDTH 80: COLOR 7,1,1: CLS: KEY OFF
100 ' print the report
110 WHILE -1
120   READ FG%, BG%, S$, NL%: COLOR FG%, BG%: PRINT S$;
130   IF NL% = 255 THEN 200
140   IF NL% THEN PRINT
150 WEND
200 ' bottom line
210 LOCATE 25,1: COLOR 1,7: PRINT "Press <Enter> to resume.";
220 COLOR 15,1: PRINT " It is recommended that you save any unsaved work.";
230 LOCATE 23,1
900 ' exit
910 COLOR 7,1: END
1000 ' template
1010 DATA 1,7,"PC-BASIC SYSTEM ERROR",0, 7,1,"",1
1020 DATA 7,1,"version   ",0, 15,1,"{version}",1
1030 DATA 7,1,"python    ",0, 15,1,"{python_version}",1
1040 DATA 7,1,"platform  ",0, 15,1,"{os_version}",1
1050 DATA 7,1,"interface ",0, 15,1,"{interface}",1
1060 DATA 7,1,"statement ",0, 15,1,"{statement}",1
1070 DATA 7,1,"",1
1080 DATA 15,1,"{traceback_0}",1
1090 DATA 15,1,"{traceback_1}",1
1100 DATA 15,1,"{traceback_2}",1
1110 DATA 15,1,"{traceback_3}",1
1120 DATA 15,1,"{exc_type}: ",0, 7,1,"{exc_value}",1
1130 DATA 7,1,"",1
1140 DATA 1,7,"This is a bug in PC-BASIC.",0, 7,1,"",1
1150 DATA 7,1,"Sorry about that. You can help improve PC-BASIC:",1
1160 DATA 7,1,"- Please file a bug report at",1
1170 DATA 15,1,"  {bug_url}",1
1180 DATA 7,1,"- Please include the full crash log stored at",1
1190 DATA 15,1,"  {crashlog}",1
1200 DATA 7,1,"",255
"""

def _bluescreen(session, iface, argv, log_dir, exc_type, exc_value, exc_traceback):
    """Process crash information and write reports."""
    # retrieve information from the defunct session
    session_info = _debrief_session(session)
    # interface information
    if iface:
        iface_name = u'%s, %s' % (type(iface._video).__name__, type(iface._audio).__name__)
    else:
        iface_name = u'--'
    # platform information
    python_version = u'%s [%s] %s' % (
        platform.python_version(), u' '.join(platform.architecture()),
        getattr(sys, 'frozen', u'') or u''
    )
    # format traceback
    # make sure the list is long enough if the traceback is not
    stack = traceback.extract_tb(exc_traceback)
    traceback_lines=[
        u'{0}:{1}, {2}'.format(os.path.basename(s[0]), s[1], s[2])
        for s in stack[-4:]
    ] + [u''] * 4
    # standard traceback format
    python_traceback=traceback.format_exception(exc_type, exc_value, exc_traceback)
    # write crash log and get its name
    log_file_name = _write_crashlog(
        log_dir, argv, python_traceback,
        **session_info
    )
    # display modal message on the interface
    do_resume = _show_report(
        iface, iface_name, python_version, session_info['code_line'],
        traceback_lines, exc_type, exc_value,
        log_file_name
    )
    return do_resume


def _debrief_session(session):
    # hide further output from defunct session
    session.attach(None)
    info = dict(
        # get session information
        repr_text_screen=session.info.repr_text_screen(),
        repr_scalars=session.info.repr_scalars(),
        repr_arrays=session.info.repr_arrays(),
        repr_strings=session.info.repr_strings(),
        repr_program=session.info.repr_program(),
        # obtain statement being executed
        code_line=session.info.get_current_code(as_type=text_type),
        # get code listing
        listing=session.execute('LIST', as_type=text_type),
    )
    return info


def _get_exception_info(exc_type, exc_value, exc_traceback):
    """Format the traceback."""
    stack = traceback.extract_tb(exc_traceback)
    exception_info = dict(
        # make sure the list is long enough if the traceback is not
        traceback_lines=[
            u'{0}:{1}, {2}'.format(os.path.basename(s[0]), s[1], s[2])
            for s in stack[-4:]
        ] + [u''] * 4,
        # standard traceback format
        python_traceback=traceback.format_exception(exc_type, exc_value, exc_traceback)
    )
    return exception_info


def _write_crashlog(
        log_dir, argv, python_traceback,
        repr_text_screen,
        repr_scalars,
        repr_arrays,
        repr_strings,
        repr_program,
        code_line,
        listing
    ):
    def separator(title):
        return u'==== {} ='.format(title).ljust(100, u'=')

    # create crash log
    crashlog = [
        u'PC-BASIC crash log',
        u'=' * 100,
        code_line, u'',
        separator('Traceback'), u''.join(python_traceback),
        separator('Version'), info.get_version_info(),
        separator('Platform'), info.get_platform_info(),
        separator('Options'), repr(argv), u'',
        separator('Screen Pages'), repr_text_screen,
        separator('Scalars'), repr_scalars,
        separator('Arrays'), repr_arrays,
        separator('Strings'), repr_strings,
        separator('Program Buffer'), repr_program, u'',
        separator('Program'), listing,
    ]
    # create crash log file
    logname = datetime.now().strftime(LOG_PATTERN)
    logfile = tempfile.NamedTemporaryFile(
        mode='wb', suffix='.log', prefix=logname, dir=log_dir, delete=False,
    )
    with logfile as f:
        # write UTF-8 Byte Order mark to ensure Notepad recognises encoding
        f.write(b'\xef\xbb\xbf')
        for line in crashlog:
            if isinstance(line, bytes): # pragma: no cover
                # for python 2 - show as cp437
                line = line.decode('cp437')
            # convert CR and CRLF line endings to CRLF, for Notepad
            f.write(line.encode('utf-8', 'replace').replace(b'\r\n', b'\n').replace(b'\n', b'\r\n'))
            f.write(b'\r\n')
    # open text file
    webbrowser.open(logfile.name)
    return logfile.name


def _show_report(iface, iface_name, python_version, code_line, traceback_lines, exc_type, exc_value, log_file_name):
    """Show a crash report on the interface."""
    resume = False
    with Session(output_streams='stdio' if not iface else ()) as session:
        # display report
        # construct the message
        message = REPORT_TEMPLATE.format(
            version=LONG_VERSION,
            python_version=python_version,
            os_version=platform.platform(),
            interface=iface_name,
            statement=code_line.replace('"', "'").replace('\n', ''),
            traceback_0=traceback_lines[0],
            traceback_1=traceback_lines[1],
            traceback_2=traceback_lines[2],
            traceback_3=traceback_lines[3],
            exc_type=u'{0}'.format(exc_type.__name__),
            exc_value=u'{0}'.format(exc_value),
            bug_url=u'https://github.com/robhagemans/pcbasic/issues',
            crashlog=log_file_name,
        )
        session.execute(message)
        session.execute('RUN')
        session.attach(iface)
    if not iface:
        return False
    while True:
        signal = iface.pause(CAPTION)
        if signal.event_type == signals.KEYB_DOWN and signal.params[0] == '\r':
            return True
        elif signal.event_type == signals.QUIT:
            return False



"""
PC-BASIC - data.codepages
Codepage definition files

(c) 2013--2023 Rob Hagemans
This file is released under the GNU GPL version 3 or later.
"""

import logging
import binascii

from ...compat import resources, unichr


# list of available codepages
CODEPAGES = tuple(
    name.split(u'.', 1)[0]
    for name in resources.contents(__package__)
    if name.lower().endswith('.ucp') and resources.is_resource(__package__, name)
)


def read_codepage(codepage_name):
    """Read a codepage file and convert to codepage dict."""
    codepage_name += '.ucp'
    codepage = {}
    for line in resources.read_binary(__package__, codepage_name).splitlines():
        # ignore empty lines and comment lines (first char is #)
        if (not line) or (line[0] == b'#'):
            continue
        # strip off comments; split unicodepoint and hex string
        splitline = line.split(b'#')[0].split(b':')
        # ignore malformed lines
        if len(splitline) < 2:
            continue
        try:
            # extract codepage point
            cp_point = binascii.unhexlify(splitline[0].strip())
            # allow sequence of code points separated by commas
            grapheme_cluster = u''.join(
                unichr(int(ucs_str.strip(), 16)) for ucs_str in splitline[1].split(b',')
            )
            codepage[cp_point] = grapheme_cluster
        except (ValueError, TypeError):
            logging.warning('Could not parse line in codepage file: %s', repr(line))
    return codepage



"""
PC-BASIC - parports.py
Parallel port handling

(c) 2013--2023 Rob Hagemans
This file is released under the GNU GPL version 3 or later.
"""

import logging
import sys
import os
import io

try:
    import parallel
except Exception:
    parallel = None

from ...compat import line_print, iterchar, stdio
from ..base import error
from ..codepage import CONTROL
from .devicebase import Device, DeviceSettings, TextFileBase, parse_protocol_string, safe_io


# flush triggers
TRIGGERS = {'page': b'\f', 'line': b'\n', 'close': None, '': None}


###############################################################################
# LPT ports

class LPTDevice(Device):
    """Parallel port or printer device (LPTn:) """

    # LPT1 can be opened as RANDOM
    # but behaves just like OUTPUT
    # in GW-BASIC, FIELD gives a FIELD OVERFLOW; we get BAD FILE MODE.
    allowed_modes = b'OR'

    def __init__(self, arg, default_stream, codepage):
        """Initialise LPTn: device."""
        Device.__init__(self)
        addr, val = parse_protocol_string(arg)
        self.stream = default_stream
        if addr == u'FILE':
            try:
                self.stream = io.open(val, 'wb')
            except EnvironmentError as e:
                logging.warning(u'Could not attach file %s to LPT device: %s', val, e)
        elif addr == u'PARPORT':
            # port can be e.g. /dev/parport0 on Linux or LPT1 on Windows. Just a number counting from 0 would also work.
            try:
                self.stream = ParallelStream(val)
            except EnvironmentError as e:
                logging.warning(u'Could not attach parallel port %s to LPT device: %s', val, e)
        elif addr == u'STDIO' or (not addr and val == u'STDIO'):
            self.stream = StdIOParallelStream()
        elif addr == u'PRINTER' or (val and not addr):
            # 'PRINTER' is default
            # name:parameters (LINE, PAGE, ...)
            options = val.split(u':')
            printer_name = options[0]
            flush_trigger = (options[1:] or [u''])[0]
            self.stream = PrinterStream(printer_name, flush_trigger, codepage)
        elif val:
            logging.warning(u'Could not attach %s to LPT device', arg)
        # column counter is the same across all LPT files
        self.device_settings = DeviceSettings()
        if self.stream:
            self.device_file = LPTFile(self.stream, self.device_settings)

    def open(
            self, number, param, filetype, mode, access, lock,
            reclen, seg, offset, length, fiekd):
        """Open a file on LPTn: """
        # shared position/width settings across files
        return LPTFile(self.stream, self.device_settings, bug=True)

    def available(self):
        """Device is available."""
        return self.stream is not None


###############################################################################
# file on LPT device

class LPTFile(TextFileBase):
    """LPTn: device - line printer or parallel port."""

    def __init__(self, stream, settings, bug=False):
        """Initialise LPTn."""
        # GW-BASIC quirk - different LPOS behaviour on LPRINT and LPT1 files
        self._bug = bug
        self._settings = settings
        TextFileBase.__init__(self, stream, filetype=b'D', mode=b'A')
        # default width is 80
        # width=255 means line wrap
        self.width = 80
        # we need to keep these in sync as self .col is accessed by Formatter (and others)
        # we can't make col a @property as the TextFileBase init tries to set it to a number
        self.col = self._settings.col

    def set_width(self, new_width=255):
        """Set file width."""
        self.width = new_width

    def write(self, s, can_break=True):
        """Write a string to the printer buffer."""
        if not isinstance(s, bytes):
            raise TypeError("LPTFile writes require bytes, not '%s'" % (type(s),))
        with safe_io():
            for c in iterchar(s):
                # don't replace CR or LF with
                self._fhandle.write(c)
                # col reverts to 1 on CR (\r) and LF (\n) but not FF (\f)
                if c in (b'\n', b'\r'):
                    self._settings.col = 1
                elif c == b'\b':
                    if self._settings.col > 1:
                        self._settings.col -= 1
                else:
                    # nonprinting characters including tabs are not counted for LPOS
                    if ord(c) >= 32:
                        self._settings.col += 1
                # width 255 means wrapping enabled
                if can_break and self.width != 255:
                    if self._settings.col > self.width:
                        self._fhandle.write(b'\r\n')
                        # GW-BASIC quirk: on LPT1 files the LPOS goes to width+1, then wraps to 2
                        if not self._bug:
                            self._settings.col = 1
                        elif self._settings.col > self.width + 1:
                            self._settings.col = 2
        self.col = self._settings.col

    def write_line(self, s=b''):
        """Write string or bytearray and newline to file."""
        assert isinstance(s, bytes), type(s)
        self.write(s + b'\r\n')

    def lof(self):
        """LOF: bad file mode """
        raise error.BASICError(error.BAD_FILE_MODE)

    def loc(self):
        """LOC: bad file mode """
        raise error.BASICError(error.BAD_FILE_MODE)

    def eof(self):
        """EOF: bad file mode """
        raise error.BASICError(error.BAD_FILE_MODE)

    def do_print(self):
        """Actually print, reset column position."""
        with safe_io():
            self._fhandle.flush()
        self._settings.col = 1
        self.col = 1

    def close(self):
        """Close the printer device and actually print the output."""
        self.do_print()


##############################################################################
# printers

class PrinterStream(io.BytesIO):
    """LPT output to printer."""

    def __init__(self, printer_name, flush_trigger, codepage):
        """Initialise the printer stream."""
        self.printer_name = printer_name
        self.codepage = codepage
        # flush_trigger can be a char or a code word
        self._flush_trigger = TRIGGERS.get(flush_trigger.lower(), flush_trigger)
        io.BytesIO.__init__(self)

    def close(self):
        """Close the printer stream."""
        self.flush()

    def write(self, s):
        """Write to printer stream."""
        for c in iterchar(s):
            if c == b'\b':
                # backspace: drop a non-newline character from the buffer
                self.seek(-1, 1)
                if self.read(1) not in (b'\r', b'\n', b'\f'):
                    self.seek(-1, 1)
                    self.truncate()
            io.BytesIO.write(self, c)
            if c == self._flush_trigger:
                self.flush()

    def flush(self):
        """Flush the printer buffer to a printer."""
        printbuf = self.getvalue()
        if not printbuf:
            return
        self.seek(0)
        self.truncate()
        # any naked lead bytes in DBCS will remain just that - avoid in-line flushes.
        utf8buf = self.codepage.bytes_to_unicode(
            printbuf, preserve=CONTROL,
        ).encode('utf-8', 'replace')
        line_print(utf8buf, self.printer_name)

    def set_control(self, select=False, init=False, lf=False, strobe=False):
        """Set the values of the control pins."""

    def get_status(self):
        """Get the values of the status pins."""
        return False, False, False, False, False


##############################################################################
# physical parallel ports

class ParallelStream(object):
    """LPT output to parallel port."""

    def __init__(self, port):
        """Initialise the ParallelStream."""
        if not parallel:
            raise IOError('`parallel` module not found. Parallel port communication not available.')
        try:
            self._parallel = parallel.Parallel(port)
        except TypeError:
            raise IOError('Invalid port specification.')
        self._port = port

    def __getstate__(self):
        """Get pickling dict for stream."""
        return {'port': self._port}

    def __setstate__(self, st):
        """Initialise stream from pickling dict."""
        self.__init__(st['port'])

    def write(self, s):
        """Write to the parallel port."""
        with safe_io():
            if self._parallel.getInPaperOut():
                raise error.BASICError(error.OUT_OF_PAPER)
            for c in iterchar(s):
                self._parallel.setData(ord(c))

    def set_control(self, select=False, init=False, lf=False, strobe=False):
        """Set the values of the control pins."""
        with safe_io():
            self._parallel.setDataStrobe(strobe)
            self._parallel.setAutoFeed(lf)
            self._parallel.setInitOut(init)
            # select-printer pin not implemented

    def get_status(self):
        """Get the values of the status pins."""
        with safe_io():
            paper = self._parallel.getInPaperOut()
            ack = self._parallel.getInAcknowledge()
            select = self._parallel.getInSelected()
        # not implemented: busy, error pins
        busy = False
        err = False
        return busy, ack, paper, select, err

    def close(self):
        """Close the stream."""
        pass


##############################################################################
# standard output

class StdIOParallelStream(object):
    """LPT output to standard output."""

    def __init__(self):
        """Initialise the stream."""

    def close(self):
        """Close the connection."""

    def write(self, s):
        """Write to stdout."""
        stdio.stdout.buffer.write(s)
        self.flush()

    def flush(self):
        """Flush stdout."""
        stdio.stdout.flush()

    def set_control(self, select=False, init=False, lf=False, strobe=False):
        """Set the values of the control pins."""

    def get_status(self):
        """Get the values of the status pins."""
        return False, False, False, False, False



"""
PC-BASIC - font.py
Font handling

(c) 2014--2023 Rob Hagemans
This file is released under the GNU GPL version 3 or later.
"""

import os
import logging
import binascii

from ...compat import iteritems, int2byte, zip

from ..base import bytematrix
from ..data import DEFAULT_FONT as _DEFAULT_FONT


# ascii codepoints for which to repeat column 8 in column 9 (box drawing)
# Many internet sources say this should be 0xC0--0xDF. However, that would
# exclude the shading characters. It appears to be traced back to a mistake in
# IBM's VGA docs. See https://01.org/linuxgraphics/sites/default/files/documentation/ilk_ihd_os_vol3_part1r2.pdf
_CARRY_COL_9_BYTES = tuple(range(0xb0, 0xdf+1))
# ascii codepoints for which to repeat row 8 in row 9 (box drawing)
_CARRY_ROW_9_BYTES = tuple(range(0xb0, 0xdf+1))


class Font(object):
    """Single-height bitfont."""

    def __init__(self, height, fontdict, codepage):
        """Initialise the font."""
        self._width = 8
        self._height = int(height)
        self._codepage = codepage
        if not fontdict:
            if height == 8:
                fontdict = {
                    self._byte_to_char(_i): binascii.unhexlify(_glyph)
                    for _i, _glyph in enumerate(_DEFAULT_FONT)
                }
            else:
                raise ValueError(
                    'No font dictionary specified and no %d-pixel default available.' % (height,)
                )
        self._fontdict = fontdict
        self._glyphs = {}
        self._carry_row_9_chars = [self._byte_to_char(_b) for _b in _CARRY_ROW_9_BYTES]
        self._carry_col_9_chars = [self._byte_to_char(_b) for _b in _CARRY_COL_9_BYTES]

    @property
    def width(self):
        return self._width

    @property
    def height(self):
        return self._height

    def copy(self):
        """Make a deep copy."""
        copy = self.__class__(self._height, self._fontdict.copy(), self._codepage)
        copy._width = self._width
        return copy

    def init_mode(self, width, height):
        """Preload SBCS glyphs at mode switch."""
        if self._width != width or self._height != height:
            self._width = width
            self._height = height
            # build the basic 256 codepage characters
            for _c in range(256):
                self._build_glyph(self._byte_to_char(_c), fullwidth=False)
        return self

    def get_byte(self, byte, offset):
        """Get byte value from character sequence."""
        char = self._byte_to_char(byte)
        return ord(self._fontdict[char][offset:offset+1])

    def set_byte(self, byte, offset, byte_value):
        """Set byte value for character sequence."""
        char = self._byte_to_char(byte)
        old = self._fontdict[char]
        self._fontdict[char] = old[:offset%8] + int2byte(byte_value) + old[offset%8+1:]
        if char in self._glyphs:
            self._build_glyph(char, fullwidth=False)

    def _byte_to_char(self, byte):
        """Map single byte value to unicode character."""
        return self._codepage.codepoint_to_unicode(int2byte(byte), use_substitutes=True)

    def _get_glyph(self, char, fullwidth):
        """Retrieve a glyph, building if needed."""
        try:
            return self._glyphs[char]
        except KeyError:
            self._build_glyph(char, fullwidth)
            return self._glyphs[char]

    def _build_glyph(self, char, fullwidth):
        """Build a glyph for the given unicode character."""
        try:
            byteseq = bytearray(self._fontdict[char])
        except KeyError:
            logging.debug('No glyph for code point %r; replacing with blank.', char)
            byteseq = bytearray(self._height)
        # shape of encoded mask (8 or 16 wide; usually 8, 14 or 16 tall)
        code_height = 8 if self._height == 9 else self._height
        if len(byteseq) < code_height:
            byteseq = byteseq.ljust(code_height, b'\0')
        glyph = bytematrix.ByteMatrix.frompacked(byteseq, code_height, items_per_byte=8)
        # stretch or sqeeze if necessary
        req_width = self._width * (2 if fullwidth else 1)
        if req_width >= glyph.width * 2:
            logging.debug('Code point %r stretched to full-width.', char)
            glyph = glyph.hrepeat(2)
        elif glyph.width >= (req_width-1) * 2:
            logging.debug('Code point %r squeezed to half-width.', char)
            glyph = glyph[:, ::2]
        # repeat last rows (e.g. for 9-bit high chars)
        if self._height > glyph.height:
            glyph = _extend_height(glyph, char in self._carry_row_9_chars)
        # repeat last cols (e.g. for 9-bit wide chars)
        if req_width > glyph.width:
            glyph = _extend_width(glyph, char in self._carry_col_9_chars)
        self._glyphs[char] = glyph

    def render_text(self, unicode_list, attr, back, underline):
        """Return a sprite, width and height for given row of text."""
        sprite = self.get_glyphs(unicode_list).render(back, attr)
        if underline:
            sprite[-1:, :] = attr
        return sprite

    def get_glyphs(self, unicode_list):
        """
        Retrieve a row of text as a single matrix [y][x].
        Text is given as list of unicode, with fullwidth characters marked by trailing u''.
        """
        # find width of each character
        # last character can't be fullwidth as it's not trailed by u''
        # note that we assign a fw value to u'' markers, but this is ignored below
        fw_list = (not _next for _next in unicode_list[1:] + [True])
        # skip u'' markers
        return bytematrix.hstack(
            self._get_glyph(_c, _fw)
            for _c, _fw in zip(unicode_list, fw_list) if _c
        )


def _extend_height(glyph, carry_last):
    """Extend the character height by a row."""
    if carry_last:
        return bytematrix.vstack((glyph, glyph[-1, :]))
    else:
        return glyph.vextend(1)

def _extend_width(glyph, carry_last):
    """Extend the character width by a column."""
    # use two empty columns if doublewidth
    if glyph.width >= 16:
        return glyph.hextend(2)
    if carry_last:
        return bytematrix.hstack((glyph, glyph[:, -1]))
    return glyph.hextend(1)



"""
PC-BASIC - video_cli.py
Command-line interface

(c) 2013--2023 Rob Hagemans
This file is released under the GNU GPL version 3 or later.
"""

import sys
import time

from .video import VideoPlugin
from .base import video_plugins, InitFailed
from ..basic.base import signals
from ..basic.base import scancode
from ..basic.base.eascii import as_unicode as uea
from ..compat import EOF, console


# escape sequence to scancode
KEY_TO_SCAN = {
    'F1': scancode.F1,  'F2': scancode.F2,  'F3': scancode.F3,  'F4': scancode.F4,
    'F5': scancode.F5,  'F6': scancode.F6,  'F7': scancode.F7,  'F8': scancode.F8,
    'F9': scancode.F9,  'F10': scancode.F10,  'F11': scancode.F11, 'F12': scancode.F12,
    'END': scancode.END,  'HOME': scancode.HOME,
    'UP': scancode.UP,  'DOWN': scancode.DOWN,  'RIGHT': scancode.RIGHT,  'LEFT': scancode.LEFT,
    'INSERT': scancode.INSERT,  'DELETE': scancode.DELETE,  'PAGEUP': scancode.PAGEUP,
    'PAGEDOWN': scancode.PAGEDOWN,
}
MOD_TO_SCAN = {
    'SHIFT': scancode.LSHIFT,
    'CTRL': scancode.CTRL,
    'ALT': scancode.ALT,
}
# escape sequence to e-ASCII
KEY_TO_EASCII = {
    'F1': uea.F1,  'F2': uea.F2,  'F3': uea.F3,  'F4': uea.F4,  'F5': uea.F5,
    'F6': uea.F6,  'F7': uea.F7,  'F8': uea.F8,  'F9': uea.F9,  'F10': uea.F10,
    'F11': uea.F11,  'F12': uea.F12,  'END': uea.END,
    'HOME': uea.HOME,  'UP': uea.UP,  'DOWN': uea.DOWN,
    'RIGHT': uea.RIGHT,  'LEFT': uea.LEFT,  'INSERT': uea.INSERT,
    'DELETE': uea.DELETE,  'PAGEUP': uea.PAGEUP,  'PAGEDOWN': uea.PAGEDOWN,
}

SHIFT_KEY_TO_EASCII = {
    'F1': uea.SHIFT_F1,
    'F2': uea.SHIFT_F2,
    'F3': uea.SHIFT_F3,
    'F4': uea.SHIFT_F4,
    'F5': uea.SHIFT_F5,
    'F6': uea.SHIFT_F6,
    'F7': uea.SHIFT_F7,
    'F8': uea.SHIFT_F8,
    'F9': uea.SHIFT_F9,
    'F10': uea.SHIFT_F10,
    'F11': uea.SHIFT_F11,
    'F12': uea.SHIFT_F12,
    'HOME': uea.SHIFT_HOME,
    'UP': uea.SHIFT_UP,
    'PAGEUP': uea.SHIFT_PAGEUP,
    'LEFT': uea.SHIFT_LEFT,
    'RIGHT': uea.SHIFT_RIGHT,
    'END': uea.SHIFT_END,
    'DOWN': uea.SHIFT_DOWN,
    'PAGEDOWN': uea.SHIFT_PAGEDOWN,
    'INSERT': uea.SHIFT_INSERT,
    'DELETE': uea.SHIFT_DELETE,
}

CTRL_KEY_TO_EASCII = {
    'F1': uea.CTRL_F1,
    'F2': uea.CTRL_F2,
    'F3': uea.CTRL_F3,
    'F4': uea.CTRL_F4,
    'F5': uea.CTRL_F5,
    'F6': uea.CTRL_F6,
    'F7': uea.CTRL_F7,
    'F8': uea.CTRL_F8,
    'F9': uea.CTRL_F9,
    'F10': uea.CTRL_F10,
    'F11': uea.CTRL_F11,
    'F12': uea.CTRL_F12,
    'HOME': uea.CTRL_HOME,
    'PAGEUP': uea.CTRL_PAGEUP,
    'LEFT': uea.CTRL_LEFT,
    'RIGHT': uea.CTRL_RIGHT,
    'END': uea.CTRL_END,
    'PAGEDOWN': uea.CTRL_PAGEDOWN,
    #'ESCAPE': uea.CTRL_ESCAPE,
    #'BACKSPACE': uea.CTRL_BACKSPACE,
    #'2': uea.CTRL_2,
    #'6': uea.CTRL_6,
    #'MINUS': uea.CTRL_MINUS,
}

ALT_KEY_TO_EASCII = {
    '1': uea.ALT_1,
    '2': uea.ALT_2,
    '3': uea.ALT_3,
    '4': uea.ALT_4,
    '5': uea.ALT_5,
    '6': uea.ALT_6,
    '7': uea.ALT_7,
    '8': uea.ALT_8,
    '9': uea.ALT_9,
    '0': uea.ALT_0,
    '-': uea.ALT_MINUS,
    '=': uea.ALT_EQUALS,
    'q': uea.ALT_q,
    'w': uea.ALT_w,
    'e': uea.ALT_e,
    'r': uea.ALT_r,
    't': uea.ALT_t,
    'y': uea.ALT_y,
    'u': uea.ALT_u,
    'i': uea.ALT_i,
    'o': uea.ALT_o,
    'p': uea.ALT_p,
    'a': uea.ALT_a,
    's': uea.ALT_s,
    'd': uea.ALT_d,
    'f': uea.ALT_f,
    'g': uea.ALT_g,
    'h': uea.ALT_h,
    'j': uea.ALT_j,
    'k': uea.ALT_k,
    'l': uea.ALT_l,
    'z': uea.ALT_z,
    'x': uea.ALT_x,
    'c': uea.ALT_c,
    'v': uea.ALT_v,
    'b': uea.ALT_b,
    'n': uea.ALT_n,
    'm': uea.ALT_m,
    '\b': uea.ALT_BACKSPACE,
    '\t': uea.ALT_TAB,
    '\r': uea.ALT_RETURN,
    ' ': uea.ALT_SPACE,
    'F1': uea.ALT_F1,
    'F2': uea.ALT_F2,
    'F3': uea.ALT_F3,
    'F4': uea.ALT_F4,
    'F5': uea.ALT_F5,
    'F6': uea.ALT_F6,
    'F7': uea.ALT_F7,
    'F8': uea.ALT_F8,
    'F9': uea.ALT_F9,
    'F10': uea.ALT_F10,
    'F11': uea.ALT_F11,
    'F12': uea.ALT_F12,
}


class VideoTextBase(VideoPlugin):
    """Text-based interface."""

    def __init__(self, input_queue, video_queue, **kwargs):
        """Initialise text-based interface."""
        if not console:
            raise InitFailed('This interface requires a console terminal (tty).')
        VideoPlugin.__init__(self, input_queue, video_queue)
        # start the stdin thread for non-blocking reads
        self._input_handler = InputHandlerCLI(input_queue)

    def _check_input(self):
        """Handle keyboard events."""
        self._input_handler.drain_queue()


@video_plugins.register('cli')
class VideoCLI(VideoTextBase):
    """Command-line interface."""

    def __init__(self, input_queue, video_queue, **kwargs):
        """Initialise command-line interface."""
        VideoTextBase.__init__(self, input_queue, video_queue)
        # current row and column where the cursor should be
        # keep cursor_row and last_row unset at the start to avoid printing extra line on resume
        # as it will see a move frm whatever we set it at here to the actusl cursor row
        self._cursor_row, self._cursor_col = 1, 1
        # current actual print column
        self._col = 1
        # cursor row on last cycle
        self._last_row = 1
        # text buffer
        self._text = [[u' '] * 80 for _ in range(25)]

    def __enter__(self):
        """Open command-line interface."""
        VideoTextBase.__enter__(self)
        console.set_raw()

    def __exit__(self, type, value, traceback):
        """Close command-line interface."""
        try:
            if self._col != 1:
                console.write(u'\r\n')
            console.unset_raw()
        finally:
            VideoTextBase.__exit__(self, type, value, traceback)


    ###############################################################################

    def update(self, row, col, unicode_matrix, attr_matrix, y0, x0, sprite):
        """Put text or pixels at a given position."""
        # if multiple rows are updated, they must be sent in order
        for ofs, unicode_list in enumerate(unicode_matrix):
            unicode_list = [(_c if _c != u'\0' else u' ') for _c in unicode_list]
            self._text[row-1+ofs][col-1:col-1+len(unicode_list)] = unicode_list
        self._refresh(row, row + ofs)

    def move_cursor(self, row, col, attr, width):
        """Move the cursor to a new position."""
        # update cursor row only if it's changed from last work-cycle
        # or if actual printing takes place on the new cursor row
        self._cursor_row, self._cursor_col = row, col
        # update cursor row only if it's changed from last work-cycle
        # or if actual printing takes place on the new cursor row
        if self._cursor_row != self._last_row or self._cursor_col != self._col:
            self._update_position(self._cursor_row, self._cursor_col)

    def clear_rows(self, back_attr, start, stop):
        """Clear screen rows."""
        self._text[start-1:stop] = [[u' '] * len(self._text[0]) for _ in range(start-1, stop)]
        self._refresh(start, stop)

    def scroll(self, direction, start_row, stop_row, back_attr):
        """Scroll the screen between start_row and stop_row."""
        if direction == -1:
            self._scroll_up(start_row, stop_row, back_attr)
        else:
            self._scroll_down(start_row, stop_row, back_attr)
        self._refresh(start_row, stop_row)

    def _scroll_up(self, start_row, stop_row, back_attr):
        """Scroll the screen up between start_row and stop_row."""
        self._text[start_row-1:stop_row] = (
            self._text[start_row:stop_row] + [[u' '] * len(self._text[0])]
        )
        if start_row < self._last_row <= stop_row:
            self._last_row -= 1

    def _scroll_down(self, start_row, stop_row, back_attr):
        """Scroll the screen down between start_row and stop_row."""
        self._text[start_row-1:stop_row] = (
            [[u' '] * len(self._text[0])] + self._text[start_row-1:stop_row-1]
        )
        if start_row <= self._last_row < stop_row:
            self._last_row += 1

    def set_mode(self, canvas_height, canvas_width, text_height, text_width):
        """Initialise video mode """
        self._text = [[u' '] * text_width for _ in range(text_height)]
        self._refresh(1, text_height)
        self._last_row = 1

    ###############################################################################

    def _refresh(self, start_row, stop_row=None):
        """Refresh the current row, if between start and stop inclusive."""
        if stop_row is None:
            stop_row = start_row
        if start_row <= self._last_row <= stop_row:
            self._redraw_row(self._last_row)
            self._redraw_row(self._last_row, self._col)

    def _update_position(self, row, col):
        """Move terminal print location."""
        # show the intermediate lines up to and including the current
        if row != self._last_row:
            if row > self._last_row:
                self._redraw_range(self._last_row, row)
            elif row == self._last_row-1:
                self._redraw_row(self._last_row-1)
        # redraw until current column to put cursor in the right position
        self._redraw_row(row, col)
        self._last_row = row
        self._col = col

    def _redraw_range(self, start_row, stop_row):
        """Redraw text for a range of rows."""
        for update_row in range(start_row, stop_row):
            self._redraw_row(update_row)
            # go one row down
            console.write(u'\n')
        self._redraw_row(stop_row)

    def _redraw_row(self, row, col=None):
        """Draw the stored text in a row."""
        rowtext = self._text[row-1]
        # we need to slice *before* joining to get the cursor right on dbcs
        # as double-width characters come through as pairs c, u''
        if col is not None:
            rowtext = rowtext[:col-1]
        # go to column 1
        console.write(u'\r')
        console.write(u''.join(rowtext))


###############################################################################

class InputHandlerCLI(object):
    """Keyboard reader thread."""

    def __init__(self, queue):
        """Start the keyboard reader."""
        self._input_queue = queue
        self._f12_active = False
        self.quit_on_eof = True

    def drain_queue(self):
        """Handle keyboard events."""
        while True:
            # s is one unicode char or one scancode
            uc, sc, mods = self._get_key()
            if not uc and not sc:
                break
            if uc == EOF and self.quit_on_eof:
                # ctrl-D (unix) / ctrl-Z (windows)
                self._input_queue.put(signals.Event(signals.QUIT))
            elif uc == u'\x7f':
                # backspace
                self._input_queue.put(
                    signals.Event(signals.KEYB_DOWN, (uea.BACKSPACE, scancode.BACKSPACE, []))
                )
            elif sc or uc:
                # check_full=False to allow pasting chunks of text
                self._input_queue.put(signals.Event(signals.KEYB_DOWN, (uc, sc, mods)))
                # this is needed since we don't send key-up events at all otherwise
                if sc == scancode.F12:
                    self._f12_active = True
                elif self._f12_active:
                    self._input_queue.put(signals.Event(signals.KEYB_UP, (scancode.F12,)))
                    self._f12_active = False

    def _get_key(self):
        """Retrieve one keypress."""
        char, key, mods = console.read_key()
        if not char and not key:
            return None, None, []
        # override characters wth alt
        if 'ALT' in mods and 'CTRL' not in mods:
            char = ALT_KEY_TO_EASCII.get(key, u'')
        # don't override ctrl characters
        if not char:
            if 'CTRL' in mods and 'ALT' not in mods:
                char = CTRL_KEY_TO_EASCII.get(key, u'')
            elif mods == {'SHIFT'}:
                char = SHIFT_KEY_TO_EASCII.get(key, char)
            else:
                char = KEY_TO_EASCII.get(key, char)
        scan = KEY_TO_SCAN.get(key, None)
        modscan = [MOD_TO_SCAN[mod] for mod in mods if mod in MOD_TO_SCAN]
        return char, scan, modscan



"""
PC-BASIC - tokens.py
BASIC keyword tokens

(c) 2014--2023 Rob Hagemans
This file is released under the GNU GPL version 3 or later.
"""

from ...compat import iteritems


# ascii constants in bytes form
# note that strings module equivanets are unicode in python 3
DIGITS = b'0123456789'
UPPERCASE = b'ABCDEFGHIJKLMNOPQRSTUVWXYZ'
LOWERCASE = UPPERCASE.lower()
LETTERS = UPPERCASE + LOWERCASE
ALPHANUMERIC = LETTERS + DIGITS
HEXDIGITS = DIGITS + b'abcdefABCDEF'
OCTDIGITS = b'01234567'


# allowable as chars 2.. in a variable name (first char must be a letter)
NAME_CHARS = ALPHANUMERIC + b'.'
# type characters
SIGILS = (b'#', b'!', b'%', b'$')

# indirect line number references
T_UINT_PROC = b'\x0d'
T_UINT = b'\x0e'
# number type tokens
T_OCT = b'\x0b'
T_HEX = b'\x0c'
T_BYTE = b'\x0f'
T_INT = b'\x1c'
T_SINGLE = b'\x1d'
T_DOUBLE = b'\x1f'
# number constants
C_0 = b'\x11'
C_1 = b'\x12'
C_2 = b'\x13'
C_3 = b'\x14'
C_4 = b'\x15'
C_5 = b'\x16'
C_6 = b'\x17'
C_7 = b'\x18'
C_8 = b'\x19'
C_9 = b'\x1a'
C_10 = b'\x1b'
# keyword tokens recognised by PC-BASIC
END = b'\x81'
FOR = b'\x82'
NEXT = b'\x83'
DATA = b'\x84'
INPUT = b'\x85'
DIM = b'\x86'
READ = b'\x87'
LET = b'\x88'
GOTO = b'\x89'
RUN = b'\x8a'
IF = b'\x8b'
RESTORE = b'\x8c'
GOSUB = b'\x8d'
RETURN = b'\x8e'
REM = b'\x8f'
STOP = b'\x90'
PRINT = b'\x91'
CLEAR = b'\x92'
LIST = b'\x93'
NEW = b'\x94'
ON = b'\x95'
WAIT = b'\x96'
DEF = b'\x97'
POKE = b'\x98'
CONT = b'\x99'
OUT = b'\x9c'
LPRINT = b'\x9d'
LLIST = b'\x9e'
WIDTH = b'\xa0'
ELSE = b'\xa1'
TRON = b'\xa2'
TROFF = b'\xa3'
SWAP = b'\xa4'
ERASE = b'\xa5'
EDIT = b'\xa6'
ERROR = b'\xa7'
RESUME = b'\xa8'
DELETE = b'\xa9'
AUTO = b'\xaa'
RENUM = b'\xab'
DEFSTR = b'\xac'
DEFINT = b'\xad'
DEFSNG = b'\xae'
DEFDBL = b'\xaf'
LINE = b'\xb0'
WHILE = b'\xb1'
WEND = b'\xb2'
CALL = b'\xb3'
WRITE = b'\xb7'
OPTION = b'\xb8'
RANDOMIZE = b'\xb9'
OPEN = b'\xba'
CLOSE = b'\xbb'
LOAD = b'\xbc'
MERGE = b'\xbd'
SAVE = b'\xbe'
COLOR = b'\xbf'
CLS = b'\xc0'
MOTOR = b'\xc1'
BSAVE = b'\xc2'
BLOAD = b'\xc3'
SOUND = b'\xc4'
BEEP = b'\xc5'
PSET = b'\xc6'
PRESET = b'\xc7'
SCREEN = b'\xc8'
KEY = b'\xc9'
LOCATE = b'\xca'
TO = b'\xcc'
THEN = b'\xcd'
TAB = b'\xce'
STEP = b'\xcf'
USR = b'\xd0'
FN = b'\xd1'
SPC = b'\xd2'
NOT = b'\xd3'
ERL = b'\xd4'
ERR = b'\xd5'
STRING = b'\xd6'
USING = b'\xd7'
INSTR = b'\xd8'
O_REM = b'\xd9'
VARPTR = b'\xda'
CSRLIN = b'\xdb'
POINT = b'\xdc'
OFF = b'\xdd'
INKEY = b'\xde'
O_GT = b'\xe6'
O_EQ = b'\xe7'
O_LT = b'\xe8'
O_PLUS = b'\xe9'
O_MINUS = b'\xea'
O_TIMES = b'\xeb'
O_DIV = b'\xec'
O_CARET = b'\xed'
AND = b'\xee'
OR = b'\xef'
XOR = b'\xf0'
EQV = b'\xf1'
IMP = b'\xf2'
MOD = b'\xf3'
O_INTDIV = b'\xf4'
CVI = b'\xfd\x81'
CVS = b'\xfd\x82'
CVD = b'\xfd\x83'
MKI = b'\xfd\x84'
MKS = b'\xfd\x85'
MKD = b'\xfd\x86'
EXTERR = b'\xfd\x8b'
FILES = b'\xfe\x81'
FIELD = b'\xfe\x82'
SYSTEM = b'\xfe\x83'
NAME = b'\xfe\x84'
LSET = b'\xfe\x85'
RSET = b'\xfe\x86'
KILL = b'\xfe\x87'
PUT = b'\xfe\x88'
GET = b'\xfe\x89'
RESET = b'\xfe\x8a'
COMMON = b'\xfe\x8b'
CHAIN = b'\xfe\x8c'
DATE = b'\xfe\x8d'
TIME = b'\xfe\x8e'
PAINT = b'\xfe\x8f'
COM = b'\xfe\x90'
CIRCLE = b'\xfe\x91'
DRAW = b'\xfe\x92'
PLAY = b'\xfe\x93'
TIMER = b'\xfe\x94'
ERDEV = b'\xfe\x95'
IOCTL = b'\xfe\x96'
CHDIR = b'\xfe\x97'
MKDIR = b'\xfe\x98'
RMDIR = b'\xfe\x99'
SHELL = b'\xfe\x9a'
ENVIRON = b'\xfe\x9b'
VIEW = b'\xfe\x9c'
WINDOW = b'\xfe\x9d'
PMAP = b'\xfe\x9e'
PALETTE = b'\xfe\x9f'
LCOPY = b'\xfe\xa0'
CALLS = b'\xfe\xa1'
PCOPY = b'\xfe\xa5'
LOCK = b'\xfe\xa7'
UNLOCK = b'\xfe\xa8'
LEFT = b'\xff\x81'
RIGHT = b'\xff\x82'
MID = b'\xff\x83'
SGN = b'\xff\x84'
INT = b'\xff\x85'
ABS = b'\xff\x86'
SQR = b'\xff\x87'
RND = b'\xff\x88'
SIN = b'\xff\x89'
LOG = b'\xff\x8a'
EXP = b'\xff\x8b'
COS = b'\xff\x8c'
TAN = b'\xff\x8d'
ATN = b'\xff\x8e'
FRE = b'\xff\x8f'
INP = b'\xff\x90'
POS = b'\xff\x91'
LEN = b'\xff\x92'
STR = b'\xff\x93'
VAL = b'\xff\x94'
ASC = b'\xff\x95'
CHR = b'\xff\x96'
PEEK = b'\xff\x97'
SPACE = b'\xff\x98'
OCT = b'\xff\x99'
HEX = b'\xff\x9a'
LPOS = b'\xff\x9b'
CINT = b'\xff\x9c'
CSNG = b'\xff\x9d'
CDBL = b'\xff\x9e'
FIX = b'\xff\x9f'
PEN = b'\xff\xa0'
STICK = b'\xff\xa1'
STRIG = b'\xff\xa2'
EOF = b'\xff\xa3'
LOC = b'\xff\xa4'
LOF = b'\xff\xa5'
# PCjr and Tandy only
NOISE = b'\xfe\xa4'
TERM = b'\xfe\xa6'

KW_END = b'END'
KW_FOR = b'FOR'
KW_NEXT = b'NEXT'
KW_DATA = b'DATA'
KW_INPUT = b'INPUT'
KW_DIM = b'DIM'
KW_READ = b'READ'
KW_LET = b'LET'
KW_GOTO = b'GOTO'
KW_RUN = b'RUN'
KW_IF = b'IF'
KW_RESTORE = b'RESTORE'
KW_GOSUB = b'GOSUB'
KW_RETURN = b'RETURN'
KW_REM = b'REM'
KW_STOP = b'STOP'
KW_PRINT = b'PRINT'
KW_CLEAR = b'CLEAR'
KW_LIST = b'LIST'
KW_NEW = b'NEW'
KW_ON = b'ON'
KW_WAIT = b'WAIT'
KW_DEF = b'DEF'
KW_POKE = b'POKE'
KW_CONT = b'CONT'
KW_OUT = b'OUT'
KW_LPRINT = b'LPRINT'
KW_LLIST = b'LLIST'
KW_WIDTH = b'WIDTH'
KW_ELSE = b'ELSE'
KW_TRON = b'TRON'
KW_TROFF = b'TROFF'
KW_SWAP = b'SWAP'
KW_ERASE = b'ERASE'
KW_EDIT = b'EDIT'
KW_ERROR = b'ERROR'
KW_RESUME = b'RESUME'
KW_DELETE = b'DELETE'
KW_AUTO = b'AUTO'
KW_RENUM = b'RENUM'
KW_DEFSTR = b'DEFSTR'
KW_DEFINT = b'DEFINT'
KW_DEFSNG = b'DEFSNG'
KW_DEFDBL = b'DEFDBL'
KW_LINE = b'LINE'
KW_WHILE = b'WHILE'
KW_WEND = b'WEND'
KW_CALL = b'CALL'
KW_WRITE = b'WRITE'
KW_OPTION = b'OPTION'
KW_RANDOMIZE = b'RANDOMIZE'
KW_OPEN = b'OPEN'
KW_CLOSE = b'CLOSE'
KW_LOAD = b'LOAD'
KW_MERGE = b'MERGE'
KW_SAVE = b'SAVE'
KW_COLOR = b'COLOR'
KW_CLS = b'CLS'
KW_MOTOR = b'MOTOR'
KW_BSAVE = b'BSAVE'
KW_BLOAD = b'BLOAD'
KW_SOUND = b'SOUND'
KW_BEEP = b'BEEP'
KW_PSET = b'PSET'
KW_PRESET = b'PRESET'
KW_SCREEN = b'SCREEN'
KW_KEY = b'KEY'
KW_LOCATE = b'LOCATE'
KW_TO = b'TO'
KW_THEN = b'THEN'
KW_TAB = b'TAB('
KW_STEP = b'STEP'
KW_USR = b'USR'
KW_FN = b'FN'
KW_SPC = b'SPC('
KW_NOT = b'NOT'
KW_ERL = b'ERL'
KW_ERR = b'ERR'
KW_STRING = b'STRING$'
KW_USING = b'USING'
KW_INSTR = b'INSTR'
KW_O_REM = b"'"
KW_VARPTR = b'VARPTR'
KW_CSRLIN = b'CSRLIN'
KW_POINT = b'POINT'
KW_OFF = b'OFF'
KW_INKEY = b'INKEY$'
KW_O_GT = b'>'
KW_O_EQ = b'='
KW_O_LT = b'<'
KW_O_PLUS = b'+'
KW_O_MINUS = b'-'
KW_O_TIMES = b'*'
KW_O_DIV = b'/'
KW_O_CARET = b'^'
KW_AND = b'AND'
KW_OR = b'OR'
KW_XOR = b'XOR'
KW_EQV = b'EQV'
KW_IMP = b'IMP'
KW_MOD = b'MOD'
KW_O_INTDIV = b'\\'
KW_CVI = b'CVI'
KW_CVS = b'CVS'
KW_CVD = b'CVD'
KW_MKI = b'MKI$'
KW_MKS = b'MKS$'
KW_MKD = b'MKD$'
KW_EXTERR = b'EXTERR'
KW_FILES = b'FILES'
KW_FIELD = b'FIELD'
KW_SYSTEM = b'SYSTEM'
KW_NAME = b'NAME'
KW_LSET = b'LSET'
KW_RSET = b'RSET'
KW_KILL = b'KILL'
KW_PUT = b'PUT'
KW_GET = b'GET'
KW_RESET = b'RESET'
KW_COMMON = b'COMMON'
KW_CHAIN = b'CHAIN'
KW_DATE = b'DATE$'
KW_TIME = b'TIME$'
KW_PAINT = b'PAINT'
KW_COM = b'COM'
KW_CIRCLE = b'CIRCLE'
KW_DRAW = b'DRAW'
KW_PLAY = b'PLAY'
KW_TIMER = b'TIMER'
KW_ERDEV = b'ERDEV'
KW_IOCTL = b'IOCTL'
KW_CHDIR = b'CHDIR'
KW_MKDIR = b'MKDIR'
KW_RMDIR = b'RMDIR'
KW_SHELL = b'SHELL'
KW_ENVIRON = b'ENVIRON'
KW_VIEW = b'VIEW'
KW_WINDOW = b'WINDOW'
KW_PMAP = b'PMAP'
KW_PALETTE = b'PALETTE'
KW_LCOPY = b'LCOPY'
KW_CALLS = b'CALLS'
KW_PCOPY = b'PCOPY'
KW_LOCK = b'LOCK'
KW_UNLOCK = b'UNLOCK'
KW_LEFT = b'LEFT$'
KW_RIGHT = b'RIGHT$'
KW_MID = b'MID$'
KW_SGN = b'SGN'
KW_INT = b'INT'
KW_ABS = b'ABS'
KW_SQR = b'SQR'
KW_RND = b'RND'
KW_SIN = b'SIN'
KW_LOG = b'LOG'
KW_EXP = b'EXP'
KW_COS = b'COS'
KW_TAN = b'TAN'
KW_ATN = b'ATN'
KW_FRE = b'FRE'
KW_INP = b'INP'
KW_POS = b'POS'
KW_LEN = b'LEN'
KW_STR = b'STR$'
KW_VAL = b'VAL'
KW_ASC = b'ASC'
KW_CHR = b'CHR$'
KW_PEEK = b'PEEK'
KW_SPACE = b'SPACE$'
KW_OCT = b'OCT$'
KW_HEX = b'HEX$'
KW_LPOS = b'LPOS'
KW_CINT = b'CINT'
KW_CSNG = b'CSNG'
KW_CDBL = b'CDBL'
KW_FIX = b'FIX'
KW_PEN = b'PEN'
KW_STICK = b'STICK'
KW_STRIG = b'STRIG'
KW_EOF = b'EOF'
KW_LOC = b'LOC'
KW_LOF = b'LOF'

KW_NOISE = b'NOISE'
KW_TERM = b'TERM'

# non-keywords that appear as syntax elements
W_AS = b'AS'
W_SHARED = b'SHARED'
W_ACCESS = b'ACCESS'
W_RANDOM = b'RANDOM'
W_OUTPUT = b'OUTPUT'
W_APPEND = b'APPEND'
W_BASE = b'BASE'
W_SEG = b'SEG'
W_ALL = b'ALL'

# other keywords on http://www.chebucto.ns.ca/~af380/GW-BASIC-tokens.html :
# Sperry PC only:
#   0xFEA4: b'DEBUG' (conflicts with PCjr/Tandy NOISE)
# Undefined tokens:
#   0x9A,  0x9B,  0x9F,  0xB4,  0xB5,  0xB6,  0xCB,  0xDF,  0xE0,  0xE1,  0xE2
#   0xE3,  0xE4,  0xE5,  0xF5,  0xF6,  0xF7,  0xF8,  0xF9,  0xFA,  0xFB,  0xFC
DIGIT = (C_0, C_1, C_2, C_3, C_4, C_5, C_6, C_7, C_8, C_9)
NUMBER = (
    T_OCT, T_HEX, T_BYTE, T_INT, T_SINGLE, T_DOUBLE,
    C_0, C_1, C_2, C_3, C_4, C_5, C_6, C_7, C_8, C_9, C_10
)
LINE_NUMBER = (T_UINT, T_UINT_PROC)
OPERATOR = (
    O_GT, O_EQ, O_LT, O_PLUS, O_MINUS,
    O_TIMES, O_DIV, O_CARET, O_INTDIV
)
# comment tokens
COMMENT = (O_REM, REM)

# line ending tokens
END_LINE = (b'\0', b'')
# statement ending tokens
END_STATEMENT = END_LINE + (b':',)
# expression ending tokens
END_EXPRESSION = END_STATEMENT + (b')', b']', b',', b';')
## tokens followed by one or more bytes to be skipped
PLUS_BYTES = {
    T_BYTE: 1, b'\xff': 1 , b'\xfe': 1, b'\xfd': 1, T_OCT: 2, T_HEX: 2,
    T_UINT_PROC: 2, T_UINT: 2, T_INT: 2, T_SINGLE: 4, T_DOUBLE: 8, b'\0': 4
}

# keyword dictionary
KEYWORDS = {
    END: KW_END, FOR: KW_FOR, NEXT: KW_NEXT, DATA: KW_DATA, INPUT: KW_INPUT,
    DIM: KW_DIM, READ: KW_READ, LET: KW_LET, GOTO: KW_GOTO, RUN: KW_RUN, IF: KW_IF,
    RESTORE: KW_RESTORE, GOSUB: KW_GOSUB, RETURN: KW_RETURN, REM: KW_REM,
    STOP: KW_STOP, PRINT: KW_PRINT, CLEAR: KW_CLEAR, LIST: KW_LIST, NEW: KW_NEW,
    ON: KW_ON, WAIT: KW_WAIT, DEF: KW_DEF, POKE: KW_POKE, CONT: KW_CONT, OUT: KW_OUT,
    LPRINT: KW_LPRINT, LLIST: KW_LLIST, WIDTH: KW_WIDTH, ELSE: KW_ELSE,
    TRON: KW_TRON, TROFF: KW_TROFF, SWAP: KW_SWAP, ERASE: KW_ERASE, EDIT: KW_EDIT,
    ERROR: KW_ERROR, RESUME: KW_RESUME, DELETE: KW_DELETE, AUTO: KW_AUTO,
    RENUM: KW_RENUM, DEFSTR: KW_DEFSTR, DEFINT: KW_DEFINT, DEFSNG: KW_DEFSNG,
    DEFDBL: KW_DEFDBL, LINE: KW_LINE, WHILE: KW_WHILE, WEND: KW_WEND, CALL: KW_CALL,
    WRITE: KW_WRITE, OPTION: KW_OPTION, RANDOMIZE: KW_RANDOMIZE, OPEN: KW_OPEN,
    CLOSE: KW_CLOSE, LOAD: KW_LOAD, MERGE: KW_MERGE, SAVE: KW_SAVE, COLOR: KW_COLOR,
    CLS: KW_CLS, MOTOR: KW_MOTOR, BSAVE: KW_BSAVE, BLOAD: KW_BLOAD, SOUND: KW_SOUND,
    BEEP: KW_BEEP, PSET: KW_PSET, PRESET: KW_PRESET, SCREEN: KW_SCREEN, KEY: KW_KEY,
    LOCATE: KW_LOCATE, TO: KW_TO, THEN: KW_THEN, TAB: KW_TAB, STEP: KW_STEP,
    USR: KW_USR, FN: KW_FN, SPC: KW_SPC, NOT: KW_NOT, ERL: KW_ERL, ERR: KW_ERR,
    STRING: KW_STRING, USING: KW_USING, INSTR: KW_INSTR, O_REM: KW_O_REM,
    VARPTR: KW_VARPTR, CSRLIN: KW_CSRLIN, POINT: KW_POINT, OFF: KW_OFF,
    INKEY: KW_INKEY, O_GT: KW_O_GT, O_EQ: KW_O_EQ, O_LT: KW_O_LT,
    O_PLUS: KW_O_PLUS, O_MINUS: KW_O_MINUS, O_TIMES: KW_O_TIMES, O_DIV: KW_O_DIV,
    O_CARET: KW_O_CARET, AND: KW_AND, OR: KW_OR, XOR: KW_XOR, EQV: KW_EQV, IMP: KW_IMP,
    MOD: KW_MOD, O_INTDIV: KW_O_INTDIV, CVI: KW_CVI, CVS: KW_CVS, CVD: KW_CVD, MKI: KW_MKI,
    MKS: KW_MKS, MKD: KW_MKD, EXTERR: KW_EXTERR, FILES: KW_FILES, FIELD: KW_FIELD,
    SYSTEM: KW_SYSTEM, NAME: KW_NAME, LSET: KW_LSET, RSET: KW_RSET, KILL: KW_KILL,
    PUT: KW_PUT, GET: KW_GET, RESET: KW_RESET, COMMON: KW_COMMON, CHAIN: KW_CHAIN,
    DATE: KW_DATE, TIME: KW_TIME, PAINT: KW_PAINT, COM: KW_COM, CIRCLE: KW_CIRCLE,
    DRAW: KW_DRAW, PLAY: KW_PLAY, TIMER: KW_TIMER, ERDEV: KW_ERDEV, IOCTL: KW_IOCTL,
    CHDIR: KW_CHDIR, MKDIR: KW_MKDIR, RMDIR: KW_RMDIR, SHELL: KW_SHELL,
    ENVIRON: KW_ENVIRON, VIEW: KW_VIEW, WINDOW: KW_WINDOW, PMAP: KW_PMAP,
    PALETTE: KW_PALETTE, LCOPY: KW_LCOPY, CALLS: KW_CALLS, PCOPY: KW_PCOPY,
    LOCK: KW_LOCK, UNLOCK: KW_UNLOCK, LEFT: KW_LEFT, RIGHT: KW_RIGHT, MID: KW_MID,
    SGN: KW_SGN, INT: KW_INT, ABS: KW_ABS, SQR: KW_SQR, RND: KW_RND, SIN: KW_SIN,
    LOG: KW_LOG, EXP: KW_EXP, COS: KW_COS, TAN: KW_TAN, ATN: KW_ATN, FRE: KW_FRE,
    INP: KW_INP, POS: KW_POS, LEN: KW_LEN, STR: KW_STR, VAL: KW_VAL, ASC: KW_ASC,
    CHR: KW_CHR, PEEK: KW_PEEK, SPACE: KW_SPACE, OCT: KW_OCT, HEX: KW_HEX,
    LPOS: KW_LPOS, CINT: KW_CINT, CSNG: KW_CSNG, CDBL: KW_CDBL, FIX: KW_FIX,
    PEN: KW_PEN, STICK: KW_STICK, STRIG: KW_STRIG, EOF: KW_EOF, LOC: KW_LOC,
    LOF: KW_LOF,
}


class TokenKeywordDict(object):
    """Token to keyword conversion for given BASIC syntax."""

    def __init__(self, syntax):
        """Build dictionaries."""
        self.to_keyword = dict(KEYWORDS)
        if syntax in ('pcjr', 'tandy'):
            # pcjr, tandy; incompatible with Sperry PC.
            self.to_keyword[NOISE] = KW_NOISE
            self.to_keyword[TERM] = KW_TERM
        self.to_token = dict((reversed(item) for item in iteritems(self.to_keyword)))
